package provide DS9 1.0

proc UpdateAnalysisMenu {} {
    global ds9
    global current
    global analysisMenu
    global analysisType

    if {$current(frame) != ""} {
	.menuBar entryconfig Analysis -state normal
	for {set i 0} {$i<$ds9(analysisItem)} {incr i} {
	    if {[$current(frame) get fits count]>0} {
		set fn [$current(frame) get fits file name number 1]
		if {[regexp "(.*)\.$analysisType($i)(.*)" $fn matched] } {
		    .menuBar.analysis entryconfig $analysisMenu($i) \
			-state normal
		} else {
		    .menuBar.analysis entryconfig $analysisMenu($i) \
			-state disabled
		}
	    } else {
		.menuBar.analysis entryconfig $analysisMenu($i) \
		    -state disabled
	    }
	}
    } else {
	.menuBar entryconfig Analysis -state disabled
    }
}

proc InitAnalysisFile {} {
    global ds9

    # look for local version
    ProcessAnalysisFile "./$ds9(extAnalysis)"

    # look for $HOME version
    ProcessAnalysisFile "~/$ds9(extAnalysis)"

    UpdateAnalysisMenu
}

proc ProcessAnalysisFile {fn} {
    global ds9
    global analysisMenu
    global analysisType
    global analysisCmd

    set status 1
    if [file exists "$fn"] {
	set id [open $fn r]
	while {[gets $id line] >= 0} {
	    # empty line
	    if {[string length $line] == 0} continue
	    # comments	    
	    if {[string range $line 0 0] == "\#"} continue
	    # else
	    switch $status {
		1 {
		    set menu "$line"
		    set template ""
		    set status 2
		}
		2 {
		    set template "$line"
		    set status 3
		}
		3 {
		    # eat the line
		    set status 4
		}
		4 {
		    if {"$menu" != "" && "$template" != "" && "$line" != ""} {
			foreach t $template {
			    set f $ds9(analysisItem)
			    set analysisMenu($f) "$menu"
			    set analysisType($f) "$t"
			    set analysisCmd($f) "$line"
			    .menuBar.analysis add command -label "$menu" \
				-command "AnalysisCmd $f"
			    incr ds9(analysisItem)
			}
		    }

		    set status 1
		}
	    }
	}
	close $id
    }
}

proc AnalysisCmd {which} {
    global analysisType
    global analysisCmd
    global current

    set str $analysisCmd($which)
    if {$current(frame) != ""} {
	if {[$current(frame) get fits count]>0} {
	    set fn [$current(frame) get fits full file name number 1]
	    regsub -all {\$filename} $str $fn result
	    set str $result

	    set str [MacroRegion $str {\$include_regions_pixels} \
		 "$current(frame) marker list pros image degrees include yes"]
	    set str [MacroRegion $str {\$include_regions_degrees} \
		 "$current(frame) marker list pros fk5 degrees include yes"]
	    set str [MacroRegion $str {\$include_regions_hms} \
		 "$current(frame) marker list pros fk5 hms include yes"]
	    set str [MacroRegion $str {\$include_regions} \
		 "$current(frame) marker list pros image degrees include yes"]

	    set str [MacroRegion $str {\$exclude_regions_pixels} \
		 "$current(frame) marker list pros image degrees include no"]
	    set str [MacroRegion $str {\$exclude_regions_degrees} \
		 "$current(frame) marker list pros fk5 degrees include no"]
	    set str [MacroRegion $str {\$exclude_regions_hms} \
		 "$current(frame) marker list pros fk5 hms include no"]
	    set str [MacroRegion $str {\$exclude_regions} \
		 "$current(frame) marker list pros image degrees include no"]

	    set str [MacroRegion $str {\$regions_pixels} \
		 "$current(frame) marker list pros image degrees none yes"]
	    set str [MacroRegion $str {\$regions_degrees} \
		 "$current(frame) marker list pros fk5 degrees none yes"]
	    set str [MacroRegion $str {\$regions_hms} \
		 "$current(frame) marker list pros fk5 hms none yes"]
	    set str [MacroRegion $str {\$regions} \
		 "$current(frame) marker list pros image degrees none yes"]
	}
    }
    
    if [catch {set ch [open "| $str"]}] {
	Error "An error as occured invoking the Analysis task."
	return
    }
    DisplayAnalysisDialog [read -nonewline $ch]
    close $ch
}

proc MacroRegion {str exp cmd } {
    # get any regions
    set region [eval $cmd]

    # we have to quote {"}, else problems down the road
    regsub -all {\"} $region {\\"} result
    set region $result

    # if we have nothing, set to none
    if {$region == ""} {
	set region "NONE"
    }

    regsub -all $exp $str "$region " result
    set str $result

    return $str
}

proc DisplayAnalysisDialog {txt} {
    global ds9
    global analysisDialog

    set mb $analysisDialog(mb)

    # see if we already have a analysisDialog window visible

    if {![winfo exist $analysisDialog(top)]} {

	set title "Analysis"

	# create the analysisDialog window

	toplevel $analysisDialog(top)
	wm title $analysisDialog(top) $title
	wm iconname $analysisDialog(top) $title
	wm group $analysisDialog(top) $ds9(top)

	$analysisDialog(top) configure -menu $mb

	# simple menus

	menu $mb -tearoff 0
	$mb add cascade -label File -menu $mb.file
	$mb add cascade -label Edit -menu $mb.edit

	menu $mb.file -tearoff 0
	$mb.file add command -label "Print..." \
	    -command [list PrintAnalysisDialog] -state disabled
	$mb.file add separator
	$mb.file add command -label "Close" -command DestroyAnalysisDialog

	menu $mb.edit -tearoff 0
	$mb.edit add command -label "Cut" -state disabled
	$mb.edit add command -label "Copy" -command CopyAnalysisDialog
	$mb.edit add command -label "Paste" -state disabled
	$mb.edit add command -label "Clear" -state disabled
	$mb.edit add separator
	$mb.edit add command -label "Select All" \
	    -command SelectAllAnalysisDialog
	$mb.edit add command -label "Select None" \
	    -command SelectNoneAnalysisDialog

	# create the text and scroll widgets
	
	set analysisDialog(text) \
	    [text $analysisDialog(top).text -height 20 -width 80 -wrap none \
		 -font courier \
		 -yscrollcommand "$analysisDialog(top).yscroll set" \
		 -xscrollcommand "$analysisDialog(top).xscroll set"]
	
	set sy [scrollbar $analysisDialog(top).yscroll \
		    -command [list $analysisDialog(text) yview] \
		    -orient vertical]
	set sx [scrollbar $analysisDialog(top).xscroll \
		    -command [list $analysisDialog(text) xview] \
		    -orient horizontal]
	grid $analysisDialog(text) $sy -sticky news
	grid $sx -stick news
	grid rowconfigure $analysisDialog(top) 0 -weight 1
	grid columnconfigure $analysisDialog(top) 0 -weight 1

	# create a special text tag for keywords

	$analysisDialog(text) tag configure keyword -font {courier} \
	    -foreground blue
	
    }

    # clear all and insert

    $analysisDialog(text) configure -state normal
    $analysisDialog(text) delete 1.0 end

    global analysis
    global current
    $analysisDialog(text) insert end "$txt"

    $analysisDialog(text) configure -state disabled
}

proc DestroyAnalysisDialog {} {
    global analysisDialog
    
    destroy $analysisDialog(top)
    destroy $analysisDialog(mb).file
    destroy $analysisDialog(mb).edit
    destroy $analysisDialog(mb)
}

proc CopyAnalysisDialog {} {
    global analysisDialog

    clipboard clear -displayof $analysisDialog(top)
    clipboard append -displayof $analysisDialog(top) \
	[selection get -displayof $analysisDialog(top)]
}

proc SelectAllAnalysisDialog {} {
    global analysisDialog

    $analysisDialog(text) tag add sel 1.0 end
}

proc SelectNoneAnalysisDialog {} {
    global analysisDialog

    $analysisDialog(text) tag remove sel 1.0 end
}

proc PrintAnalysisDialog {} {
    global analysisDialog

    Error "Sorry, Not Implemented."
}
