package provide DS9 1.0

proc ButtonButton {button text cmd} {
    button $button \
	-text $text \
	-command $cmd \
	-takefocus 0 \
	-bd 1 \
	-height 1 \
	-padx 2 -pady 2

    pack $button -side left -fill x -expand true
}

proc RadioButton {button text var value cmd} {
    radiobutton $button \
	-text $text \
	-variable $var \
	-value $value \
	-command $cmd \
	-indicatoron false \
	-selectcolor "" \
	-takefocus 0 \
	-bd 1 \
	-height 1 \
	-padx 2 -pady 3

    pack $button -side left -fill x -expand true
}

proc CheckButton {button text var cmd} {
    checkbutton $button \
	-text $text \
	-variable $var \
	-command $cmd \
	-indicatoron false \
	-selectcolor "" \
	-takefocus 0 \
	-bd 1 \
	-height 1 \
	-padx 2 -pady 3

    pack $button -side left -fill x -expand true
}

proc CreateButtons {} {
    global buttons

    set buttons(name) [frame .ds9.buttons]

    # Major buttons

    frame $buttons(name).major
    pack $buttons(name).major -side top -fill x -expand true

    RadioButton $buttons(name).major.file File buttons(majorCurrent) \
	$buttons(name).file MajorButton
    RadioButton $buttons(name).major.frame Frame buttons(majorCurrent) \
	$buttons(name).frame MajorButton
    RadioButton $buttons(name).major.block Block buttons(majorCurrent) \
	$buttons(name).block MajorButton
    RadioButton $buttons(name).major.zoom Zoom buttons(majorCurrent) \
	$buttons(name).zoom MajorButton
    RadioButton $buttons(name).major.scale Scale buttons(majorCurrent) \
	$buttons(name).scale MajorButton
    RadioButton $buttons(name).major.color1 Color buttons(majorCurrent) \
	$buttons(name).color1 MajorButton
    RadioButton $buttons(name).major.region Region buttons(majorCurrent) \
	$buttons(name).region1 MajorButton

    FileButtons
    FrameButtons
    BlockButtons
    ZoomButtons
    ScaleButtons
    ColorButtons
    RegionButtons

    set buttons(majorPrev) ""
    set buttons(majorCurrent) $buttons(name).file
    MajorButton
}

proc MajorButton {} {
    global buttons

    pack forget $buttons(majorPrev)
    pack $buttons(majorCurrent) -side top -fill x -expand true
    set buttons(majorPrev) $buttons(majorCurrent)
}

proc FileButtons {} {
    global buttons

    frame $buttons(name).file

    ButtonButton $buttons(name).file.open open OpenFile
    ButtonButton $buttons(name).file.clear clear ClearFrame
    ButtonButton $buttons(name).file.header header DisplayHeaderMenu
    ButtonButton $buttons(name).file.print print PrintDialog
    ButtonButton $buttons(name).file.exit exit QuitDS9
}

proc FrameButtons {} {
    global buttons
    global ds9

    frame $buttons(name).frame

    ButtonButton $buttons(name).frame.new new CreateFrame
    ButtonButton $buttons(name).frame.delete delete DeleteFrameMenu
    ButtonButton $buttons(name).frame.reset reset ResetFrame
    RadioButton $buttons(name).frame.single single ds9(displaymode) single DisplayMode
    RadioButton $buttons(name).frame.tile tile ds9(displaymode) tile DisplayMode
    RadioButton $buttons(name).frame.blink blink ds9(displaymode) blink DisplayMode
    ButtonButton $buttons(name).frame.first first FirstFrame
    ButtonButton $buttons(name).frame.prev prev PrevFrame
    ButtonButton $buttons(name).frame.next next NextFrame
    ButtonButton $buttons(name).frame.last last LastFrame
}

proc BlockButtons {} {
    global buttons

    frame $buttons(name).block

    ButtonButton $buttons(name).block.in in "Block .5"
    ButtonButton $buttons(name).block.out out "Block 2"
    RadioButton $buttons(name).block.b1 "block 1" block(factor) 1 ChangeBlockFactor
    RadioButton $buttons(name).block.b2 "block 2" block(factor) 2 ChangeBlockFactor
    RadioButton $buttons(name).block.b4 "block 4" block(factor) 4 ChangeBlockFactor
    RadioButton $buttons(name).block.b8 "block 8" block(factor) 8 ChangeBlockFactor
    RadioButton $buttons(name).block.b16 "block 16" block(factor) 16 ChangeBlockFactor
    RadioButton $buttons(name).block.b32 "block 32" block(factor) 32 ChangeBlockFactor
    RadioButton $buttons(name).block.b64 "block 64" block(factor) 64 ChangeBlockFactor
    RadioButton $buttons(name).block.b128 "block 128" block(factor) 128 ChangeBlockFactor
}

proc ZoomButtons {} {
    global buttons
    global wcs

    frame $buttons(name).zoom

    ButtonButton $buttons(name).zoom.in "in" "Zoom 2"
    ButtonButton $buttons(name).zoom.out "out" "Zoom .5"
    ButtonButton $buttons(name).zoom.center "center" CenterFrame

    RadioButton $buttons(name).zoom.none none current(orient) none ChangeOrient
    RadioButton $buttons(name).zoom.x x current(orient) x ChangeOrient
    RadioButton $buttons(name).zoom.y y current(orient) y ChangeOrient
    RadioButton $buttons(name).zoom.xy xy current(orient) xy ChangeOrient

    CheckButton $buttons(name).zoom.align "align wcs" wcs(align) AlignWCSFrame
}

proc ScaleButtons {} {
    global buttons
    global scale

    frame $buttons(name).scale

    RadioButton $buttons(name).scale.minmax minmax scale(mode) minmax ScaleMode
    RadioButton $buttons(name).scale.zscale zscale scale(mode) zscale ScaleMode

    RadioButton $buttons(name).scale.linear linear current(scale) linear ChangeColorScale
    RadioButton $buttons(name).scale.log log current(scale) log ChangeColorScale
    RadioButton $buttons(name).scale.squared squared current(scale) \
	squared ChangeColorScale
    RadioButton $buttons(name).scale.sqrt sqrt current(scale) sqrt ChangeColorScale
}

proc ColorButtons {} {
    global buttons
    global colorbar

    set ids [colorbar list id]

    frame $buttons(name).color1

    CheckButton $buttons(name).color1.invert invert colorbar(invert) InvertColorbar
    for {set i 0} {$i<9} {incr i} {
	set id [lindex $ids $i]
	set title [colorbar get name $id]
	set name [string tolower $title]
	RadioButton $buttons(name).color1.$name $name \
	    colorbar(map) $title "ChangeColormapID $id"
    }
    RadioButton $buttons(name).color1.cmap cmap buttons(majorCurrent) \
	$buttons(name).color2 MajorButton

    frame $buttons(name).color2

    CheckButton $buttons(name).color2.invert invert colorbar(invert) InvertColorbar
    for {set i 9} {$i<17} {incr i} {
	set id [lindex $ids $i]
	set title [colorbar get name $id]
	set name [string tolower $title]
	RadioButton $buttons(name).color2.$name $name \
	    colorbar(map) $title "ChangeColormapID $id"
    }
    RadioButton $buttons(name).color2.cmap cmap buttons(majorCurrent) \
	$buttons(name).color1 MajorButton
}

proc RegionButtons {} {
    global buttons

    frame $buttons(name).region1

    RadioButton $buttons(name).region1.circle circle marker(shape) circle ""
    RadioButton $buttons(name).region1.ellipse ellipse marker(shape) ellipse ""
    RadioButton $buttons(name).region1.rectangle rectangle marker(shape) rectangle ""
    RadioButton $buttons(name).region1.polygon polygon marker(shape) polygon ""
    RadioButton $buttons(name).region1.line line marker(shape) line ""
    RadioButton $buttons(name).region1.text text marker(shape) text ""
    RadioButton $buttons(name).region1.ruler ruler marker(shape) ruler ""
    RadioButton $buttons(name).region1.point point marker(shape) "circle point" ""

    RadioButton $buttons(name).region1.region region buttons(majorCurrent) \
	$buttons(name).region2 MajorButton

    frame $buttons(name).region2

    ButtonButton $buttons(name).region2.info info InfoMarker
    ButtonButton $buttons(name).region2.front front FrontMarker
    ButtonButton $buttons(name).region2.back back BackMarker
    ButtonButton $buttons(name).region2.all all SelectAllMarker
    ButtonButton $buttons(name).region2.none none UnselectAllMarker
    ButtonButton $buttons(name).region2.load load OpenMarkerFile
    ButtonButton $buttons(name).region2.save save SaveMarker
    ButtonButton $buttons(name).region2.list list DisplayMarkerList

    RadioButton $buttons(name).region2.region region buttons(majorCurrent) \
	$buttons(name).region1 MajorButton
    
}



