package provide DS9 1.0

# Open

proc OpenFile {} {
    global ds9
    LoadFits [OpenFileDialog $ds9(top) ds9(lastFitsName)] ""
}

proc OpenMosaicFile {} {
    global ds9
    LoadFits [OpenFileDialog $ds9(top) ds9(lastMosaicName)] mosaic
}

proc OpenURL {} {
    global ds9
    LoadFits [URLDialog ds9(lastURL)] ""
}

proc OpenMosaicURL {} {
    global ds9
    LoadFits [URLDialog ds9(lastURL)] mosaic
}

proc OpenMosaicExtFile {} {
    global ds9
    LoadMosaicExtFile [OpenFileDialog $ds9(top) ds9(lastMosaicName)]
}

proc OpenMosaicFiles {} {
    global ds9
    LoadMosaicFiles [OpenFileDialog $ds9(top) ds9(lastMosaicName)]
}

# Load

proc LoadFits {url mode} {
    if {$url != ""} {
	if {! [regexp -nocase {^([A-Za-z]+)?(:)?(/)*([^/:]+)(:([0-9]+))?(/.*)?$} \
		   $url x proto ask slash host y port srvurl]} {
	    Error "Unable to parse URL $url"
	    return
	}

	switch $proto {
	    http {LoadHTTP $url $mode}
	    ftp {LoadFTP $url $mode}
	    file {LoadFile "$slash$host$y$port$srvurl" $mode}
	    default {LoadFile $url $mode}
	}
    }
}

proc LoadFile {fileName mode} {
    global current

    StartLoad

    if {$fileName != ""} {
	global loadParam

	set loadParam(type) mmap
	set loadParam(mode) $mode
	set loadParam(filetype) [CalcFileType $fileName]
	ConvertFile $fileName
	ProcessLoad

	UpdateHeader
	ResetFrameColorbar
    }

    FinishLoad
}

proc LoadFTP {url mode} {
    global current

    StartLoad

    regexp -nocase {^([A-Za-z]+)?(://)?([^/:]+)(:([0-9]+))?(/.*)?$} $url \
	x proto dummy host y port srvurl

    if [FTP::Open $host "ftp" "ds9@[info hostname]"] {
	set tmp "/tmp/[file tail $url]"
	FTP::Type binary
	if [FTP::Get $srvurl $tmp] {
	    global loadParam

	    # alloc it because we are going to delete it after load
	    set loadParam(type) alloc
	    set loadParam(mode) $mode
	    set loadParam(filetype) [CalcFileType $tmp]
	    ConvertFile $tmp
	    ProcessLoad

	    UpdateHeader
	    ResetFrameColorbar
	}
	if [file exists $tmp] {
	    catch {file delete -force $tmp}
	}
	FTP::Close
    }

    FinishLoad
}

proc LoadHTTP {url mode} {
    global current
    global current
    global loadParam
    global http

    StartLoad

    set tmp "/tmp/[file tail $url]"
    set ch [open $tmp w]

    set token [http::geturl $url -channel $ch]
    if {$http(log)} {
	upvar 0 $token t
	DisplayLog "$t(http)\n"
	DisplayLog "$t(type)\n"
	DisplayLog "$t(totalsize)\n"
	DisplayLog [BreakUp $t(meta)]
	DisplayLog "$t(status)\n"
    }

    # alloc it because we are going to delete it after load
    set loadParam(type) alloc
    set loadParam(mode) $mode
    set loadParam(filetype) [CalcFileType $tmp]
    ConvertFile $tmp
    ProcessLoad

    catch {close $ch}

    UpdateHeader
    ResetFrameColorbar

    if [file exists $tmp] {
	catch {file delete -force $tmp}
    }

    FinishLoad
}

proc LoadShared {id size mode filter} {
    global loadParam
    global current

    StartLoad

    set loadParam(mode) $mode
    set loadParam(id) $id
    set loadParam(size) $size
    set loadParam(name) $filter
    set loadParam(filetype) [CalcFileType $filter]
    ProcessLoad

    UpdateHeader
    ResetFrameColorbar

    FinishLoad
}

proc LoadMosaicExtFile {fileName} {
    global current
    global loadParam

    StartLoad

    if {$fileName != ""} {
	$current(frame) unload

	set loadParam(type) mmap
	set loadParam(mode) "mosaic ext"
	set loadParam(filetype) fits
	ConvertFile $fileName
	ProcessLoad

	UpdateHeader
	ResetFrameColorbar
    }

    FinishLoad
}

proc LoadMosaicFiles {fileName} {
    global current
    global loadParam
    global mosaicTypes

    StartLoad

    if {$fileName != ""} {
	$current(frame) unload

        foreach t $mosaicTypes {
            set fileNames [$t $fileName]
            if {$fileNames != ""} {
		break
	    }
        }
        if {$fileNames == ""} { 
	    set fileNames $fileName 
	}

        foreach fileName $fileNames {
	    set loadParam(type) mmap
	    set loadParam(mode) mosaic
	    set loadParam(filetype) fits
	    ConvertFile $fileName
	    ProcessLoad
        }

	UpdateHeader
	ResetFrameColorbar
    }

    FinishLoad
}

# Support

proc BreakUp {str} {
    set r ""
    set l [string length $str]
    for {set i 0} {$i < $l} {incr i} {
	set c [string index $str $i]
	append r $c
	if {$c=="\}"} {
	    append r "\n"
	}
    }
    return $r
}

proc URLDialog {prevURL} {
    upvar $prevURL result

    if {[EntryDialog "URL" "Enter World Wide Web Location (URL):" 80 result]} {
	return $result
    } else {
	return ""
    }
}

# Header

proc UpdateHeader {} {
    global current
    global header

    if [winfo exist $header(top)] {
	DestroyHeader

	set count [$current(frame) get fits count]
	switch $count {
	    0 {return}
	    default {DisplayHeader $count}
	}
    }
}

proc DisplayHeaderMenu {} {
    global current

    set count [$current(frame) get fits count]
    switch $count {
	0 {return}
	1 {set which 1}
	default {
	    set which [SimpleListBox ".headerSelect" "Select Header" \
			   [$current(frame) get fits file names]]
	    if {$which==0} {
		return
	    }
	}
    }
    DisplayHeader $which
}

proc DisplayHeader {which} {
    global ds9
    global header
    global current

    # see if we already have a header window visible

    if [winfo exist $header(top)] {
	DestroyHeader
    }

    set title [$current(frame) get fits file name number $which]

    # create the header window

    toplevel $header(top)
    wm title $header(top) $title
    wm iconname $header(top) $title
    wm group $header(top) $ds9(top)

    $header(top) configure -menu $header(mb)

    # create the text and scroll widgets
    
    set t [text $header(top).text -height 40 -width 80 -wrap none \
	       -font courier \
	       -yscrollcommand "$header(top).yscroll set" \
	       -xscrollcommand "$header(top).xscroll set"]
    
    set sy [scrollbar $header(top).yscroll -command [list $t yview] \
		-orient vertical]
    set sx [scrollbar $header(top).xscroll -command [list $t xview] \
		-orient horizontal]
    grid $t $sy -sticky news
    grid $sx -stick news
    grid rowconfigure $header(top) 0 -weight 1
    grid columnconfigure $header(top) 0 -weight 1

    # create a special text tag for keywords

    $t tag configure keyword -font {courier} -foreground blue
    
    # insert the fits header into the text widget

    $t insert end [$current(frame) get fits header $which]

    # color tag keywords

    set stop [$t index end]
    for {set i 1.0} {$i<$stop} {set i [expr $i+1]} {
	$t tag add keyword $i "$i +8 chars"
    }

    $t configure -state disabled
}

proc DestroyHeader {} {
    global header
    
    destroy $header(top)
}

proc CopyHeader {} {
    global header

    clipboard clear -displayof $header(top)
    clipboard append -displayof $header(top) [selection get -displayof $header(top)]
}

proc PrintHeader {} {
    global header

    tk_messageBox -message "Sorry, Not Implemented." \
	-parent $header(top) -type ok -icon info
}

proc UpdateFitsMenu {} {
    global current
    global buttons

    if {$current(frame) != ""} {
	.menuBar.file entryconfig "Open..." -state normal
	.menuBar.file entryconfig "Open URL..." -state normal
	.menuBar.file entryconfig "Open Mosaic Image File..." -state normal
	.menuBar.file entryconfig "Open Mosaic Image Multiple Files..." -state normal
	.menuBar.file entryconfig "Open Mosaic Segment..." -state normal
	.menuBar.file entryconfig "Open Mosaic Segment URL..." -state normal

	$buttons(name).file.open configure -state normal

	set count [$current(frame) get fits count]
	if {$count == 0} {
	    .menuBar.file entryconfig "Clear" -state disabled
	    .menuBar.file entryconfig "Display Fits Header" -state disabled

	    $buttons(name).file.clear configure -state disabled
	    $buttons(name).file.header configure -state disabled
	} else {
	    .menuBar.file entryconfig "Clear" -state normal
	    .menuBar.file entryconfig "Display Fits Header" -state normal

	    $buttons(name).file.clear configure -state normal
	    $buttons(name).file.header configure -state normal
	}

	.menuBar.frame entryconfig "Match Frames" -state normal

	if {[$current(frame) has wcs equatorial]} {
	    .menuBar.frame.match entryconfig "WCS" -state normal
	} else {
	    .menuBar.frame.match entryconfig "WCS" -state disabled
	}

	if {[$current(frame) has mosaic] && [$current(frame) has physical]} {
	    .menuBar.frame.match entryconfig "Detector" -state disabled
	} else {
	    .menuBar.frame.match entryconfig "Detector" -state disabled
	}

	if {[$current(frame) has mosaic]} {
	    .menuBar.frame.match entryconfig "Amplifier" -state disabled
	} else {
	    .menuBar.frame.match entryconfig "Amplifier" -state disabled
	}

	if {[$current(frame) has physical]} {
	    .menuBar.frame.match entryconfig "Physical" -state disabled
	} else {
	    .menuBar.frame.match entryconfig "Physical" -state disabled
	}

	.menuBar.frame entryconfig "Lock Crosshairs" -state normal

	if {[$current(frame) has wcs equatorial]} {
	    .menuBar.frame.crosshair entryconfig "WCS" -state normal
	} else {
	    .menuBar.frame.crosshair entryconfig "WCS" -state disabled
	}

	if {[$current(frame) has mosaic] && [$current(frame) has physical]} {
	    .menuBar.frame.crosshair entryconfig "Detector" -state normal
	} else {
	    .menuBar.frame.crosshair entryconfig "Detector" -state disabled
	}

	if {[$current(frame) has mosaic]} {
	    .menuBar.frame.crosshair entryconfig "Amplifier" -state normal
	} else {
	    .menuBar.frame.crosshair entryconfig "Amplifier" -state disabled
	}

	if {[$current(frame) has physical]} {
	    .menuBar.frame.crosshair entryconfig "Physical" -state normal
	} else {
	    .menuBar.frame.crosshair entryconfig "Physical" -state disabled
	}

    } else {
	.menuBar.file entryconfig "Open..." -state disabled
	.menuBar.file entryconfig "Open URL..." -state disabled
	.menuBar.file entryconfig "Open Mosaic Image File..." -state disabled
	.menuBar.file entryconfig "Open Mosaic Image Multiple Files..." \
	    -state disabled
	.menuBar.file entryconfig "Open Mosaic Segment..." -state disabled
	.menuBar.file entryconfig "Open Mosaic Segment URL..." -state disabled
	$buttons(name).file.open configure -state normal

	.menuBar.file entryconfig "Clear" -state disabled
	.menuBar.file entryconfig "Display Fits Header" -state disabled
	$buttons(name).file.clear configure -state disabled
	$buttons(name).file.header configure -state disabled

	.menuBar.frame entryconfig "Match Frames" -state disabled
	.menuBar.frame entryconfig "Lock Crosshairs" -state disabled
    }
}

proc ProcessLoad {} {
    global current
    global loadParam

    switch $loadParam(type) {
	mmap {
	    if [catch {$current(frame) load $loadParam(filetype) $loadParam(mode) \
			   mmap $loadParam(name)}] {
		Error "Unable to load $loadParam(filetype) $loadParam(mode) Image $loadParam(name)"
	    }
	}
	alloc {
	    if [catch {$current(frame) load $loadParam(filetype) $loadParam(mode) \
			   alloc $loadParam(name)}] {
		Error "Unable to load $loadParam(filetype) $loadParam(mode) Image $loadParam(name)" 
	    }
	}
	shared {
	    if [catch {$current(frame) load $loadParam(filetype) $loadParam(mode) \
			   shared $loadParam(id) $loadParam(size) $loadParam(name)}] {
		Error "Unable to load shared $loadParam(filetype) $loadParam(mode) Image $loadParam(id)"
	    }
	}
	channel {
	    fconfigure $loadParam(channel) -translation binary
	    if [catch {$current(frame) load $loadParam(filetype) $loadParam(mode) \
			   channel $loadParam(channel) $loadParam(name)}] {
		Error "Unable to read $loadParam(filetype) $loadParam(mode) Image from channel $loadParam(channel)"
	    }
	    catch {close $loadParam(channel)}
	}
    }
}

proc StartLoad {} {
    SetWatchCursor
}

proc FinishLoad {} {
    global current

    UpdateInfoFileName $current(frame)

    UpdateGraphYAxis
    UpdateGraphXAxis

    UpdateFitsMenu
    UpdateBlockMenu
    UpdateContourMenu
    UpdateWCSMenu
    UpdateAnalysisMenu

    UnsetWatchCursor
}

proc CalcFileType {fn} {
    set f [string tolower $fn]
    if {[string first "dim" $f] > 0 && [string first "bitpix" $f]} {
	return array
    } else {
	return fits
    }
}

proc ConvertFile fileName {
    global ds9
    global loadParam

    # default

    set loadParam(name) $fileName

    foreach t $ds9(fileProcs) {
	if [$t $fileName] {
	    return
	}
    }
}

# File Types

proc StdinFits {fileName} {
    global loadParam

    if {$fileName == "-" || $fileName == "stdin" || $fileName == "STDIN"} {
	set loadParam(type) alloc
	set loadParam(name) "stdin"
	return 1
    }

    return 0
}

proc CompressFits {fileName} {
    global loadParam

    if { [regexp {(.*)\.Z} $fileName matched root] } {
	if [catch {set ch [open "| uncompress < $root.Z " r]}] {
	    return 0
	}
	set loadParam(type) channel
	set loadParam(channel) $ch
	set loadParam(name) "$fileName"
	return 1
    }

    return 0
}

proc PackFits {fileName} {
    global loadParam

    if { [regexp {(.*)\.z} $fileName matched root] } {
	if [catch {set ch [open "| pcat $root.z " r]}] {
	    return 0
	}
	set loadParam(type) channel
	set loadParam(channel) $ch
	set loadParam(name) "$fileName"
	return 1
    }

    return 0
}

proc GzipFits {fileName} {
    global loadParam

    if { [regexp {(.*)\.gz} $fileName matched root] } {
	if [catch {set ch [open "| gunzip < $root.gz " r]}] {
	    return 0
	}
	set loadParam(type) channel
	set loadParam(channel) $ch
	set loadParam(name) "$fileName"
	return 1
    }

    return 0
}

proc ExternalFits {fileName} {
    global externalType
    global externalCmd
    global loadParam

    foreach a [array names externalType] {
	if { [regexp "(.*)\.$externalType($a)(.*)" $fileName matched] } {
	    regsub -all {\$filename} "$externalCmd($a)" "$matched" result
	    set cmd "| $result"
	    if [catch {set ch [open "$cmd" r]}] {
		return 0
	    }
	    set loadParam(type) channel
	    set loadParam(channel) $ch
	    set loadParam(name) "$fileName"
	    return 1
	}
    }
    return 0
}

# Mosaic File filter procedures

set mosaicTypes {
    mosaicMega
    mosaicGeneric
    mosaicGeneric2
}

proc mosaicGeneric {fileName} {
    if { [regexp {(.*)\_[0-9]+\.fits} $fileName matched root] } {
	return "[glob -nocomplain ${root}_*\.fits ]"
    } else {
	return ""
    }
}

proc mosaicGeneric2 {fileName} {
    if { [regexp {(.*)[0-9]+\.fits} $fileName matched root] } {
	return "[glob -nocomplain ${root}*\.fits ]"
    } else {
	return ""
    }
}

proc mosaicMega { fileName } {
    if { [regexp {(.*)\.IM[0-9]+\.fits} $fileName matched root] } {
	return "[glob -nocomplain ${root}.IM*\.fits]"
    } else {
	return ""
    }
}
