package provide DS9 1.0

# Public Procedures

proc InitCanvas {} {
    global canvas

    # Bindings

    bind $canvas(name) <Configure> [list LayoutFrames]
    bind $canvas(name) <Tab> [list NextFrame]
    bind $canvas(name) <Shift-Tab> [list PrevFrame]
    bind $canvas(name) <Enter> [list focus $canvas(name)]
    bind $canvas(name) <Leave> [list focus {}]

    # backward compatible bindings

    bind $canvas(name) <Button-3> {Button3Canvas %x %y}
    bind $canvas(name) <B3-Motion> {Motion3Canvas %x %y}
    bind $canvas(name) <ButtonRelease-3> {Release3Canvas %x %y}
}

proc Button3Canvas {x y} {
    global ds9

    set ds9(b3) 1
    ButtonColormap $x $y
}

proc Motion3Canvas {x y} {
    MotionColormap $x $y
}

proc Release3Canvas {x y} {
    global ds9

    set ds9(b3) 0
    ReleaseColormap $x $y
}

proc CreateFrame {} {
    CreateNamedFrame Frame
}

proc CreateNamedFrame {name} {
    global ds9

    incr ds9(sequence)
    CreateNameNumberFrame "$name$ds9(sequence)"
}

proc CreateNameNumberFrame {which} {
    global ds9
    global active
    global current
    global canvas
    global pan
    global mag
    global block
    global contour
    global minmax
    global scale
    global user

    set ds9(next) "$which"
    lappend ds9(frames) $ds9(next)
    lappend ds9(active) $ds9(next)
    set active($ds9(next)) 1

    $canvas(name) create frame$ds9(visual)$ds9(depth) \
	-x 0 -y 0 -anchor nw -command "$ds9(next)" -tag "$ds9(next)"

    $ds9(next) colormap [colorbar get colormap]
    $ds9(next) set name $ds9(next)
    $ds9(next) panner 'panner' $pan(size) $pan(size)
    $ds9(next) magnifier 'magnifier' $mag(size) $mag(size)
    $ds9(next) magnifier mode $mag(mode)
    $ds9(next) use square pixels $current(squarePixels)

    $ds9(next) colorscale $current(scale)

    $ds9(next) clip scope $scale(scope)
    $ds9(next) clip mode $scale(mode)
    $ds9(next) clip minmax mode $minmax(mode)
    $ds9(next) clip user $user(low) $user(high)

    $ds9(next) block mode manual
    $ds9(next) block function $block(function)
    $ds9(next) block factor to $block(factor)
    $ds9(next) block buffer size $block(buffersize)

    if {$ds9(mode) == "crosshair"} {
	$ds9(next) crosshair on
    }

    # add menu items

    .menuBar.frame add radiobutton -label "$ds9(next)" \
	-variable ds9(next) -value $ds9(next) -command GotoFrame
    .menuBar.frame.active add checkbutton -label "$ds9(next)" \
	-variable active($ds9(next)) -command UpdateActiveFrames

    set current(frame) $ds9(next)

    SetWatchCursor
    LayoutFrames
    UnsetWatchCursor
}

proc DeleteFrameMenu {} {
    global buttons

    if {[tk_messageBox -type okcancel -default cancel \
	     -message "Delete this Frame?" -icon question] == "ok"} {
	DeleteCurrentFrame
    }
    # reset the buttons
    set buttons(majorCurrent) $buttons(name).file
    MajorButton
}

proc DeleteCurrentFrame {} {
    global current

    if {$current(frame) != ""} {
	DeleteFrame $current(frame)
    }
}

proc DeleteFrame {which} {
    global ds9
    global active
    global header
    global current
    global canvas

    # kill the header dialog
    if {[winfo exist $header(top)]} {
	DestroyHeader
    }

    # delete any slave frames
    DeleteExamine $which

    # delete canvas widget
    $canvas(name) delete $which

    # delete menu entries
    .menuBar.frame delete "$which"
    .menuBar.frame.active delete "$which"

    # delete it from the frame list
    set i [lsearch $ds9(frames) $which]
    set ds9(frames) [lreplace $ds9(frames) $i $i]
    unset active($which)

    # and update layout
    UpdateActiveFrames
}

proc UpdateFrame {} {
    global current

    SetWatchCursor
    if {$current(frame) != ""} {
	$current(frame) update
    }
    UnsetWatchCursor
}

# Event Processing

proc BindEvents {which} {
    global canvas

    $canvas(name) bind $which <Motion> [list MotionFrame $which %x %y]
    $canvas(name) bind $which <Shift-Motion> \
	[list ShiftMotionFrame $which %x %y]
    $canvas(name) bind $which <Enter> [list EnterFrame $which %x %y]
    $canvas(name) bind $which <Leave> [list LeaveFrame $which]

    $canvas(name) bind $which <Button-1> [list Button1Frame $which %x %y]
    $canvas(name) bind $which <Shift-Button-1> \
	[list ShiftButton1Frame $which %x %y]
    $canvas(name) bind $which <B1-Motion> [list Motion1Frame $which %x %y]
    $canvas(name) bind $which <ButtonRelease-1> \
	[list Release1Frame $which %x %y]

    $canvas(name) bind $which <Double-1> [list Double1Frame $which %x %y]
    $canvas(name) bind $which <Double-ButtonRelease-1> \
	[list DoubleRelease1Frame $which %x %y]

    $canvas(name) bind $which <Button-2> [list Button2Frame $which %x %y]
    $canvas(name) bind $which <Shift-Button-2> \
	[list ShiftButton2Frame $which %x %y]
    $canvas(name) bind $which <B2-Motion> [list Motion2Frame $which %x %y]
    $canvas(name) bind $which <ButtonRelease-2> \
	[list Release2Frame $which %x %y]

    $canvas(name) bind $which <Delete> [list DeleteKeyFrame $which %x %y]
    $canvas(name) bind $which <BackSpace> [list DeleteKeyFrame $which %x %y]

    $canvas(name) bind $which <Up> [list ArrowKeyFrame $which 0 -1]
    $canvas(name) bind $which <Down> [list ArrowKeyFrame $which 0 1]
    $canvas(name) bind $which <Left> [list ArrowKeyFrame $which -1 0]
    $canvas(name) bind $which <Right> [list ArrowKeyFrame $which 1 0]
}

proc UnBindEvents {which} {
    global canvas

    $canvas(name) bind $which <Motion> {}
    $canvas(name) bind $which <Shift-Motion> {}
    $canvas(name) bind $which <Enter> {}
    $canvas(name) bind $which <Leave> {}

    $canvas(name) bind $which <Button-1> {}
    $canvas(name) bind $which <Shift-Button-1> {}
    $canvas(name) bind $which <B1-Motion> {}
    $canvas(name) bind $which <ButtonRelease-1> {}

    $canvas(name) bind $which <Double-1> {}
    $canvas(name) bind $which <Double-ButtonRelease-1> {}

    $canvas(name) bind $which <Button-2> {}
    $canvas(name) bind $which <Shift-Button-2> {}
    $canvas(name) bind $which <B2-Motion> {}
    $canvas(name) bind $which <ButtonRelease-2> {}

    $canvas(name) bind $which <Delete> {}
    $canvas(name) bind $which <BackSpace> {}

    $canvas(name) bind $which <Up> {}
    $canvas(name) bind $which <Down> {}
    $canvas(name) bind $which <Left> {}
    $canvas(name) bind $which <Right> {}
}

proc EnterFrame {which x y} {
    global ds9
    global canvas
    global info
    global mag

    # check to see if this event was generated while processing other events

    if {$ds9(b1) || $ds9(sb1) || $ds9(b2) || $ds9(b3)} {
	return
    }

    if {$ds9(centerpixel)} {
	set x [expr $x+.5]
	set y [expr $y+.5]
    }

    $canvas(name) focus $which

    # Display Filename and Coord info

    switch $ds9(mode) {
	crosshair {EnterFrameCrosshair $which}
	pointer {UpdateInfoBox $which $x $y canvas}
	colorbar -
	pan -
	zoom -
	rotate -
	examine -
	illustrate {}
    }

    # Display Frame Info

    if {[string range $which 0 4]=="Frame"} {
	set info(pFrame) "Frame [string range $which 5 end]"
    } else {
	set info(pFrame) "Frame [string range $which 7 end]"
    }
    set info(pZoom) [$which get zoom fixed]
    set info(pAngle) [$which get rotate fixed]

    # and Other stuff

    UpdateGraphData

    if {$mag(view)} {
	$which magnifier on
    }
}

proc LeaveFrame {which} {
    global ds9
    global canvas
    global info
    global mag

    # check to see if this event was generated while processing other events

    if {$ds9(b1) || $ds9(sb1) || $ds9(b2) || $ds9(b3)} {
	return
    }

    $canvas(name) focus {}

    switch $ds9(mode) {
	crosshair {}
	pointer -
	colorbar -
	pan -
	zoom -
	rotate -
	examine -
	illustrate {
	    ClearInfoBox
	    ClearGraphData
	}
    }

    switch $ds9(displaymode) {
	single {}
	tile {
	    set info(pFrame) Frame
	    set info(pZoom) ""
	    set info(pAngle) ""
	}
	blink {}
    }

    $which magnifier off
    magnifier clear
}

proc MotionFrame {which x y} {
    global ds9
    global current
    global mag

    # if button 3 is down, ignore this event, we are doing something already

    if {$ds9(b3) || $ds9(b2)} {
	return
    }

    if {$ds9(centerpixel)} {
	set x [expr $x+.5]
	set y [expr $y+.5]
    }

    switch $ds9(mode) {
	pointer {
	    if {$which == $current(frame)} {
		CursorMarker $which $x $y sizing
	    }
	    UpdateInfoBox $which $x $y canvas
	    MotionGraphCanvas $which $x $y
	}
	crosshair {}
	colorbar -
	pan -
	zoom -
	rotate -
	examine -
	illustrate
	{
	    UpdateInfoBox $which $x $y canvas
	    MotionGraphCanvas $which $x $y
	}
    }

    switch $ds9(displaymode) {
	single -
	tile {
	    if {$mag(view)} {
		$which magnifier update $x $y
	    }
	}
	blink {}
    }
}

proc ShiftMotionFrame {which x y} {
    global ds9
    global current
    global mag

    # if button 3 is down, ignore this event, we are doing something already

    if {$ds9(b3)} {
	return
    }

    if {$ds9(centerpixel)} {
	set x [expr $x+.5]
	set y [expr $y+.5]
    }

    switch $ds9(mode) {
	pointer {
	    if {$which == $current(frame)} {
		CursorMarker $which $x $y exchange
	    }
	    UpdateInfoBox $which $x $y canvas
	    MotionGraphCanvas $which $x $y
	}
	crosshair {}
	colorbar -
	pan -
	zoom -
	rotate -
	examine -
	illustrate
	{
	    UpdateInfoBox $which $x $y canvas
	    MotionGraphCanvas $which $x $y
	}
    }

    switch $ds9(displaymode) {
	single -
	tile {
	    if {$mag(view)} {
		$which magnifier update $x $y
	    }
	}
	blink {}
    }
}

proc Button1Frame {which x y} {
    global ds9
    global current
    global marker

    global debug
    if {$debug(events)} {
	puts "Button 1"
    }

    # let others know that the mouse is down

    set ds9(b1) 1

    set ax $x
    set ay $y
    if {$ds9(centerpixel)} {
	set x [expr $x+.5]
	set y [expr $y+.5]
    }

    switch $ds9(mode) {
	pointer {
	    if {$which == $current(frame)} {
		ButtonMarker $which $x $y
	    } else {
		# we need this cause MotionMarker maybe called, 
		# and we don't want it
		set marker(motion) none
		set marker(handle) -1

		set ds9(next) $which
		GotoFrame
	    }
	}
	crosshair {ButtonCrosshair $which $x $y}
	colorbar {}
	pan {ButtonPan $which $ax $ay}
	zoom {ButtonZoom $which $ax $ay}
	rotate {ButtonRotate $which $ax $ay}
	examine {ButtonExamine $which $ax $ay}
	illustrate {}
    }
}

proc ShiftButton1Frame {which x y} {
    global ds9
    global current

    # let others know that the mouse is down

    set ds9(sb1) 1
    if {$ds9(centerpixel)} {
	set x [expr $x+.5]
	set y [expr $y+.5]
    }

    switch $ds9(mode) {
	pointer {
	    if {$which == $current(frame)} {
		ShiftMarker $which $x $y
	    }
	}
	crosshair {}
	colorbar {}
	pan {}
	zoom {ShiftZoom $which}
	rotate {}
	examine {}
	illustrate {}
    }
}

proc Motion1Frame {which x y} {
    global ds9
    global current
    global mag

    global debug
    if {$debug(events)} {
	puts "Motion 1"
    }

    # abort if we are here by accident (such as a double click)

    if {($ds9(b1) == 0) && ($ds9(sb1) == 0)} {
	return
    }

    set ax $x
    set ay $y
    if {$ds9(centerpixel)} {
	set x [expr $x+.5]
	set y [expr $y+.5]
    }

    switch $ds9(mode) {
	pointer {
	    if {$which == $current(frame)} {
		MotionMarker $which $x $y
	    }
	}
	crosshair {MotionCrosshair $which $x $y}
	colorbar {}
	pan {MotionPan $which $ax $ay}
	zoom {}
	rotate {MotionRotate $which $ax $ay}
	examine {}
	illustrate {}
    }

    switch $ds9(displaymode) {
	single -
	tile {
	    if {$mag(view)} {
		$which magnifier update $x $y
	    }
	}
	blink {}
    }
}

proc Release1Frame {which x y} {
    global ds9
    global current

    global debug
    if {$debug(events)} {
	puts "Release 1"
    }

    # abort if we are here by accident (such as a double click)

    if {($ds9(b1) == 0) && ($ds9(sb1) == 0)} {
	return
    }

    set ax $x
    set ay $y
    if {$ds9(centerpixel)} {
	set x [expr $x+.5]
	set y [expr $y+.5]
    }

    switch $ds9(mode) {
	pointer {
	    if {$which == $current(frame)} {
		ReleaseMarker $which $x $y
	    }
	}
	crosshair {ReleaseCrosshair $which $x $y}
	colorbar {}
	pan {ReleasePan $which $ax $ay}
	zoom {}
	rotate {ReleaseRotate $which $ax $ay}
	examine {}
	illustrate {}
    }

    # let others know that the mouse is up

    set ds9(b1) 0
    set ds9(sb1) 0
}

proc Double1Frame {which x y} {
    global ds9
    global current

    global debug
    if {$debug(events)} {
	puts "DoubleButton 1"
    }

    set ax $x
    set ay $y
    if {$ds9(centerpixel)} {
	set x [expr $x+.5]
	set y [expr $y+.5]
    }

    switch $ds9(mode) {
	pointer {
	    if {$which == $current(frame)} {
		DoubleMarker $which $x $y
	    }
	}
	crosshair {}
	colorbar {}
	pan {}
	zoom {}
	rotate {}
	examine {}
	illustrate {}
    }
}

proc DoubleRelease1Frame {which x y} {
    global ds9

    global debug
    if {$debug(events)} {
	puts "DoubleRelease 1"
    }

    if {$ds9(centerpixel)} {
	set x [expr $x+.5]
	set y [expr $y+.5]
    }

    switch $ds9(mode) {
	pointer {}
	crosshair {}
	colorbar {}
	pan {}
	zoom {}
	rotate {}
	examine {}
	illustrate {}
    }
}

proc Button2Frame {which x y} {
    global ds9

    set ds9(b2) 1
    ButtonPan $which $x $y
}

proc ShiftButton2Frame {which x y} {
    global ds9

    set ds9(sb2) 1
}

proc Motion2Frame {which x y} {
    MotionPan $which $x $y
}

proc Release2Frame {which x y} {
    global ds9

    ReleasePan $which $x $y

    # let others know that the mouse is up

    set ds9(b2) 0
    set ds9(sb2) 0
}

proc DeleteKeyFrame {which x y} {
    global ds9

    if {$ds9(centerpixel)} {
	set x [expr $x+.5]
	set y [expr $y+.5]
    }

    switch $ds9(mode) {
	pointer {DeleteMarker $which $x $y}
	crosshair {}
	colorbar {}
	pan {}
	zoom {}
	rotate {}
	examine {}
	illustrate {}
    }
}

proc ArrowKeyFrame {which x y} {
    global ds9

    switch $ds9(mode) {
	pointer {
	    $which warp $x $y
	    $which marker move $x $y
	}
	pan -
	zoom -
	rotate {$which warp $x $y}
	crosshair {ArrowKeyCrosshair $which $x $y}
	colorbar {}
	examine {}
	illustrate {}
    }
}

# Other Public Routines

proc FirstFrame {} {
    global ds9
    set ds9(next) [lindex $ds9(active) 0]
    GotoFrame
}

proc PrevFrame {} {
    global ds9
    global current

    set i [lsearch $ds9(active) $current(frame)]
    if {$i>0} {
	set ds9(next) [lindex $ds9(active) [expr $i-1]]
    } else {
	set last [expr [llength $ds9(active)] -1]
	set ds9(next) [lindex $ds9(active) $last]
    }
    GotoFrame
}

proc NextFrame {} {
    global ds9
    global current

    set i [lsearch $ds9(active) $current(frame)]
    set last [expr [llength $ds9(active)] -1]
    if {$i<$last} {
	set ds9(next) [lindex $ds9(active) [expr $i+1]]
    } else {
	set ds9(next) [lindex $ds9(active) 0]
    }
    GotoFrame
}

proc LastFrame {} {
    global ds9
    set ds9(next) [lindex $ds9(active) [expr [llength $ds9(active)]-1]]
    GotoFrame
}

proc UpdateActiveFrames {} {
    global ds9
    global active
    global current

    # reset active list

    set ds9(active) ""
    foreach f $ds9(frames) {
	if {$active($f)} {
	    lappend ds9(active) $f
	    .menuBar.frame entryconfig $f -state normal
	} else {
	    .menuBar.frame entryconfig $f -state disabled
	}

    }

    # New layout if needed

    if {[llength $ds9(active)] > 0} {
	if {[lsearch $ds9(active) $current(frame)] == -1} {
	    set current(frame) [lindex $ds9(active) 0]
	    set ds9(next) $current(frame)
	}
    }

    SetWatchCursor
    LayoutFrames
    UnsetWatchCursor
}

proc ActiveFrameAll {} {
    global ds9
    global active

    foreach f $ds9(frames) {
	set active($f) 1
    }
    UpdateActiveFrames
}

proc ActiveFrameNone {} {
    global ds9
    global active

    foreach f $ds9(frames) {
	set active($f) 0
    }
    UpdateActiveFrames
}

proc GotoFrame {} {
    global ds9
    global current
    global active

    if {$current(frame) != "" && $current(frame) != $ds9(next)} {
	$current(frame) marker unselect all
	$current(frame) highlite off
	$current(frame) panner off

	switch $ds9(displaymode) {
	    single {$current(frame) hide}
	    tile {}
	    blink {$current(frame) hide}
	}
    }

    if {$current(frame) != $ds9(next)} {
	set current(frame) $ds9(next)
	FrameToFront $current(frame)
    }
}

proc DisplayMode {} {
    global ds9
    global blink

    switch $ds9(displaymode) {
	single -
	tile {
	    # turn off blink if on
	    if {$blink(id)>0} {
		after cancel $blink(id)
		set blink(id) 0
		set blink(index) -1
	    }

	    SetWatchCursor
	    LayoutFrames
	    UnsetWatchCursor
	}
	blink {
	    # ignore if we are already blinking
	    if {$blink(id)==0} {
		SetWatchCursor
		LayoutFrames
		UnsetWatchCursor

		BlinkTimer
	    }
	}
    }
}

proc BlinkTimer {} {
    global blink
    global ds9

    if {$ds9(active) != ""} {
	incr blink(index)
	if {$blink(index) >= [llength $ds9(active)]} {
	    set blink(index) 0
	}
	set ds9(next) [lindex $ds9(active) $blink(index)]
	GotoFrame
    }

    set blink(id) [after $blink(interval) BlinkTimer]
}

proc ResetFrame {} {
    global current

    if {$current(frame) != ""} {
	SetWatchCursor
	$current(frame) reset
	UpdateLockFrames reset

	ClearInfoBox
	ClearGraphData

	UnsetWatchCursor

	UpdatePanZoomMenu
	UpdateScaleMenu

	UpdateGraphYAxis
	UpdateGraphXAxis
    }
}

proc ClearFrame {} {
    global current
    global header

    if {$current(frame) != ""} {

	# delete any slave frames
	DeleteExamine $current(frame)

	if {[winfo exist $header(top)]} {
	    DestroyHeader
	}

	SetWatchCursor
	$current(frame) clear
	UnsetWatchCursor

	ClearInfoBox
	UpdateFrameMenus
	ClearGraphData
    }
}

proc MatchFramesImage {} {
    global ds9
    global current

    if {$current(frame) != ""} {
	SetWatchCursor

	set pan [$current(frame) get pan image scientific]

	foreach f $ds9(frames) {
	    if {$f != $current(frame)} {
		$f pan to $pan image

		$f zoom to $current(zoom)
		$f orient $current(orient)
		$f rotate to $current(rotate)
	    }
	}

	UnsetWatchCursor
    }
}

proc MatchFramesWCS {} {
    global ds9
    global current
    global wcs

    if {$current(frame) != ""} {
	SetWatchCursor

	set wcs(align) 1
	$current(frame) wcs align $wcs(align)

	set wcsZoom [$current(frame) get wcs zoom scientific]
	set wcsSystem [$current(frame) get wcs system]
	set wcsPan [$current(frame) get pan $wcsSystem scientific]

	foreach f $ds9(frames) {
	    if {$f != $current(frame)} {
		if  {[$f has wcs equatorial]} {
		    $f wcs align $wcs(align)
		    $f wcs zoom $wcsZoom

		    $f pan to $wcsPan $wcsSystem
		    $f zoom to $current(zoom)
		    $f orient $current(orient)
		    $f rotate to $current(rotate)
		}
	    }
	}

	UnsetWatchCursor
    }
}

proc ChangeSquarePixels {} {
    global ds9
    global current

    SetWatchCursor
    foreach f $ds9(frames) {
	$f use square pixels $current(squarePixels)
    }
    UnsetWatchCursor
}

proc ChangeScanOnLoad {} {
    global ds9

    SetWatchCursor
    foreach f $ds9(frames) {
	$f scan on load $ds9(scanOnLoad)
    }
    UnsetWatchCursor
}

# Private Procedures

proc UpdateFrameMenus {} {
    UpdateFrameMenu
    UpdateFitsMenu
    UpdatePanZoomMenu
    UpdateScaleMenu
    UpdateBlockMenu
    UpdateMarkerMenu
    UpdateContourMenu
    UpdateWCSMenu
    UpdateAnalysisMenu

    UpdateGraphYAxis
    UpdateGraphXAxis
}

proc UpdateLockFrames {command} {
    global ds9
    global current
    global wcs

    if {$ds9(lock)} {
	foreach f $ds9(frames) {
	    if {$f != $current(frame)} {
		eval $f $command
	    }
	}
    }
}

proc ResetFrameColorbar {} {
    global current

    UpdatePanZoomMenu
    UpdateScaleMenu

    UpdateGraphYAxis
    UpdateGraphXAxis
}

proc FrameToFront {which} {
    global ds9
    global current
    global canvas
    global colorbar
    global pan
    global info

    global debug
    if {$debug(layout)} {
	puts "FrameToFront"
    }

    colorbar set colormap [$which get colormap]
    set colorbar(map) [colorbar get name]
    set colorbar(invert) [colorbar get invert]

    $canvas(name) raise $which
    $which show
    switch $ds9(displaymode) {
	single {}
	tile {
	    if {[llength $ds9(active)] > 1} {
		$which highlite on
	    }
	}
	blink {}
    }

    if {$pan(view)} {
	$which panner on
    }
    UpdateFrameMenus

    UpdateInfoFileName $which
    if {[string range $which 0 4]=="Frame"} {
	set info(pFrame) "Frame [string range $which 5 end]"
    } else {
	set info(pFrame) "Frame [string range $which 7 end]"
    }
    set info(pZoom) [$which get zoom fixed]
    set info(pAngle) [$which get rotate fixed]
}

proc LayoutFrames {} {
    global ds9
    global current
    global canvas
    global tile
    global pan
    global mag

    global debug
    if {$debug(layout)} {
	puts "LayoutFrames"
    }

    set canvas(width) [expr [winfo width $canvas(name)] - 4]
    set canvas(height) [expr [winfo height $canvas(name)] - 4]

    # turn everything off

    foreach f $ds9(frames) {
	$f hide
	$f highlite off
	$f panner off
	$f magnifier off
	UnBindEvents $f
    }

    set n [llength $ds9(active)]
    if {$n > 0} {
	switch $ds9(displaymode) {
	    single {TileOne}
	    tile {
		switch $tile(mode) {
		    column {TileRect 1 $n}
		    row {TileRect $n 1}
		    grid {TileRect [expr int(sqrt($n-1))+1] \
			      [expr int(sqrt($n)+.5)]}
		}
	    }
	    blink {TileOne}
	}
    } else {
	set current(frame) ""
	set ds9(next) ""
	if {$pan(view)} {
	    panner clear
	}
	if {$mag(view)} {
	    magnifier clear
	}

	UpdateFrameMenus
    }
}

# This procedure is called when we have only 1 frames to display
# Canvas Coordinate system is 1 to canvas(width), canvas(height)

proc TileOne {} {
    global ds9
    global current
    global canvas

    global debug
    if {$debug(layout)} {
	puts "TileOne"
    }

    set w $canvas(width)
    set h $canvas(height)

    set x [expr int($canvas(width)/2.) + 2]
    set y [expr int($canvas(height)/2.) + 2]

    foreach f $ds9(active) {
	$f configure -x $x -y $y -width $w -height $h -anchor center
	$f show
	BindEvents $f
    }

    FrameToFront $current(frame)
}

proc TileRect {numx numy} {
    global ds9
    global canvas

    global debug
    if {$debug(layout)} {
	puts "TileRect"
    }

    set w [expr $canvas(width) /$numx - $ds9(framegap)]
    set h [expr $canvas(height)/$numy - $ds9(framegap)]
    
    for {set i 0; set b 1} {$i < $numy} {incr i; incr b 2} {
	for {set j 0; set a 1} {$j < $numx} {incr j; incr a 2} {
	    set c [expr $i*$numx + $j]
	    set x($c) [expr int($canvas(width) *$a/(2.*$numx)) + 2]
	    set y($c) [expr int($canvas(height)*$b/(2.*$numy)) + 2]
	}
    }

    TileIt $w $h x y
}

proc TileIt {w h xx yy} {
    upvar $xx x
    upvar $yy y
    global ds9
    global current
    global canvas

    global debug
    if {$debug(layout)} {
	puts "TileIt"
    }

    set i 0
    foreach f $ds9(active) {
	$f configure -x $x($i) -y $y($i) -width $w -height $h -anchor center
	$f show
	BindEvents $f
	incr i
    }

    FrameToFront $current(frame)
}

proc UpdateFrameMenu {} {
    global ds9
    global current
    global buttons

    if {$current(frame) != ""} {
	.menuBar.frame entryconfig "Delete Frame" -state normal
	.menuBar.frame entryconfig "Reset Frame" -state normal
	.menuBar.frame entryconfig "Refresh Frame" -state normal
	.menuBar.frame entryconfig "Single Frame" -state normal
	.menuBar.frame entryconfig "Tile Frames" -state normal
	.menuBar.frame entryconfig "Blink Frames" -state normal
	#	.menuBar.frame entryconfig "Lock Frames" -state normal
	.menuBar.frame entryconfig "Match Colorbars" -state normal
	#	.menuBar.frame entryconfig "Lock Colorbars" -state disabled

	$buttons(name).frame.delete configure -state normal
	$buttons(name).frame.reset configure -state normal
	$buttons(name).frame.single configure -state normal
	$buttons(name).frame.tile configure -state normal
	$buttons(name).frame.blink configure -state normal
	$buttons(name).frame.first configure -state normal
	$buttons(name).frame.prev configure -state normal
	$buttons(name).frame.next configure -state normal
	$buttons(name).frame.last configure -state normal
    } else {
	.menuBar.frame entryconfig "Delete Frame" -state disabled
	.menuBar.frame entryconfig "Reset Frame" -state disabled
	.menuBar.frame entryconfig "Refresh Frame" -state disabled
	.menuBar.frame entryconfig "Single Frame" -state disabled
	.menuBar.frame entryconfig "Tile Frames" -state disabled
	.menuBar.frame entryconfig "Blink Frames" -state disabled
	#	.menuBar.frame entryconfig "Lock Frames" -state disabled
	.menuBar.frame entryconfig "Match Colorbars" -state disabled
	#	.menuBar.frame entryconfig "Lock Colorbars" -state disabled

	$buttons(name).frame.delete configure -state disabled
	$buttons(name).frame.reset configure -state disabled
	$buttons(name).frame.single configure -state disabled
	$buttons(name).frame.tile configure -state disabled
	$buttons(name).frame.blink configure -state disabled
	$buttons(name).frame.first configure -state disabled
	$buttons(name).frame.prev configure -state disabled
	$buttons(name).frame.next configure -state disabled
	$buttons(name).frame.last configure -state disabled
    }
}
