package provide DS9 1.0

proc CenterFrame {} {
    global current

    if {$current(frame) != ""} {
	SetWatchCursor
	$current(frame) center
	UpdateLockFrames [list center]
	UpdateGraphXAxis
	UnsetWatchCursor

	UpdatePanZoomMenu
    }
}

proc Pan {x y sys} {
    global current

    if {$current(frame) != ""} {
	if {$sys == ""} {
	    set sys image
	}
	SetWatchCursor
	$current(frame) pan $x $y $sys
	UpdateGraphXAxis
	UnsetWatchCursor
    }
}

proc PanTo {x y sys} {
    global current

    if {$current(frame) != ""} {
	if {$sys == ""} {
	    set sys image
	}
	SetWatchCursor
	$current(frame) pan to $x $y $sys
	UpdateGraphXAxis
	UnsetWatchCursor
    }
}

proc ButtonPan {which x y} {
    global pan

    switch $pan(mode) {
	click {}
	drag {
	    $which pan motion begin $x $y
	    UpdateLockFrames "pan motion begin $x $y"
	}
    }
}

proc MotionPan {which x y} {
    global pan

    switch $pan(mode) {
	click {}
	drag {
	    $which pan motion $x $y
	    UpdateLockFrames "pan motion $x $y"
	}
    }
}

proc ReleasePan {which x y} {
    global pan

    switch $pan(mode) {
	click {
	    SetWatchCursor
	    $which pan to $x $y
	    UpdateLockFrames "pan to $x $y"
	}
	drag {
	    $which pan motion end $x $y

	    # we need to delay setting the cursor here, since set cursor causes
	    # an update, so the display will flash

	    SetWatchCursor
	    UpdateLockFrames "pan motion end $x $y"
	}
    }
    UpdateGraphXAxis
    UnsetWatchCursor
}

proc ZoomToFit {} {
    global current

    if {$current(frame) != ""} {
	SetWatchCursor
	$current(frame) zoom to fit
	set current(zoom) [$current(frame) get zoom]
	UpdateGraphXAxis
	UnsetWatchCursor
    }
}

proc ChangeZoom {} {
    global current

    if {$current(frame) != ""} {
	SetWatchCursor
	$current(frame) zoom to $current(zoom)
	UpdateLockFrames [list zoom to $current(zoom)]
	UpdateGraphXAxis
	UnsetWatchCursor
    }
}

proc Zoom {value} {
    global current

    if {$current(frame) != ""} {
	SetWatchCursor
	$current(frame) zoom $value
	set current(zoom) [$current(frame) get zoom]
	UpdateLockFrames [list zoom $value]
	UpdateGraphXAxis
	UnsetWatchCursor
    }
}

proc ButtonZoom {which x y} {
    global current

    if {$current(frame) != ""} {
	SetWatchCursor
	$which zoom 2 about $x $y
	if {$which == $current(frame)} {
	    set current(zoom) [$current(frame) get zoom]
	}
	UpdateLockFrames "zoom 2 about $x $y canvas"
	UpdateGraphXAxis
	UnsetWatchCursor
    }
}

proc ShiftZoom {which} {
    global current

    if {$current(frame) != ""} {
	SetWatchCursor
	$which zoom .5
	UpdateLockFrames [list zoom .5]
	if {$which == $current(frame)} {
	    set current(zoom) [$current(frame) get zoom]
	}
	UpdateGraphXAxis
	UnsetWatchCursor
    }
}

proc ChangeOrient {} {
    global current

    if {$current(frame) != ""} {
	SetWatchCursor
	$current(frame) orient $current(orient)
	UpdateLockFrames [list orient $current(orient)]
	UpdateGraphXAxis
	UnsetWatchCursor
    }
}

proc Rotate {value} {
    global current

    if {$current(frame) != ""} {
	SetWatchCursor
	$current(frame) rotate $value
	set current(rotate) [$current(frame) get rotate]
	UpdateLockFrames [list rotate $current(rotate)]
	UpdateGraphXAxis
	UnsetWatchCursor
    }
}

proc ChangeRotate {} {
    global current

    if {$current(frame) != ""} {
	SetWatchCursor
	$current(frame) rotate to $current(rotate)
	UpdateLockFrames [list rotate to $current(rotate)]
	UpdateGraphXAxis
	UnsetWatchCursor
    }
}

proc ButtonRotate {which x y} {
    global rotate

    $which rotate motion begin
    set rotate(x) $x
    UpdateLockFrames "rotate motion begin $x"
}

proc MotionRotate {which x y} {
    global rotate
    global cursor
    global angleNameValue

    $which rotate motion [expr double($rotate(x)-$x)/$cursor(speed) * 180.]
    set angleNameValue [$which get rotate fixed]
    UpdateLockFrames "rotate motion $x $y"
}

proc ReleaseRotate {which x y} {
    global angleNameValue
    global current

    $which rotate motion end
    set angleNameValue [$which get rotate fixed]
    set current(rotate) [$which get rotate fixed]

    # we need to delay setting the cursor here, since set cursor causes
    # an update, so the display will flash

    SetWatchCursor
    UpdateLockFrames "rotate motion end $x $y"
    UpdateGraphXAxis
    UnsetWatchCursor
}

proc PanZoomDialog {} {
    global panZoom
    global panZoomWindow
    global menu

    set w ".panzoom"
    set panZoom(ok) 0

    global current
    set panZoom(zoom) [$current(frame) get zoom fixed]

    set panZoom(coord) image
    set coord [$current(frame) get cursor $panZoom(coord)]
    set panZoom(x) [lindex $coord 0]
    set panZoom(y) [lindex $coord 1]

    set panZoom(rotate) [$current(frame) get rotate fixed]

    if [DialogCreate $w "Change Pan Zoom Rotate" -borderwidth 2] {
	frame $w.zoom  -relief groove -borderwidth 2
	frame $w.rotate  -relief groove -borderwidth 2
	frame $w.pan -relief groove -borderwidth 2
	frame $w.buttons -relief groove -borderwidth 2
	pack $w.zoom $w.rotate $w.pan $w.buttons -fill x
	pack $w.zoom $w.rotate $w.buttons -ipadx 4 -ipady 4

	label $w.zoom.title -text "Zoom"
	entry $w.zoom.value -textvariable panZoom(zoom) -width 9 -relief sunken -bd 2
	pack $w.zoom.title $w.zoom.value -side left -padx 4
	pack $w.zoom.value -ipadx 2 -ipady 2

	label $w.rotate.title -text "Rotate To"
	entry $w.rotate.value -textvariable panZoom(rotate) -width 9 -relief sunken -bd 2
	label $w.rotate.title2 -text "degrees"
	pack $w.rotate.title $w.rotate.value $w.rotate.title2 -side left -padx 4
	pack $w.rotate.value -ipadx 2 -ipady 2

	label $w.pan.title -text "Pan To"

	frame $w.pan.value
	entry $w.pan.value.x -textvariable panZoom(x) -width 20 -relief sunken -bd 2
	entry $w.pan.value.y -textvariable panZoom(y) -width 20 -relief sunken -bd 2
	pack $w.pan.value.x $w.pan.value.y -side top -pady 2 -ipadx 1 -ipady 1

	frame $w.pan.coord
	radiobutton $w.pan.coord.fk4 -text "Equatorial B1950" \
	    -variable panZoom(coord) -value fk4 -selectcolor $menu(selectcolor)
	radiobutton $w.pan.coord.fk5 -text "Equatorial J2000" \
	    -variable panZoom(coord) -value fk5 -selectcolor $menu(selectcolor)
	radiobutton $w.pan.coord.icrs -text "ICRS" \
	    -variable panZoom(coord) -value icrs -selectcolor $menu(selectcolor)
	radiobutton $w.pan.coord.galactic -text "Galactic" \
	    -variable panZoom(coord) -value galactic -selectcolor $menu(selectcolor)
	radiobutton $w.pan.coord.ecliptic -text "Ecliptic" \
	    -variable panZoom(coord) -value ecliptic -selectcolor $menu(selectcolor)
	radiobutton $w.pan.coord.detector -text "Detector" \
	    -variable panZoom(coord) -value detector -selectcolor $menu(selectcolor)
	radiobutton $w.pan.coord.amplifier -text "Amplifier" \
	    -variable panZoom(coord) -value amplifier -selectcolor $menu(selectcolor)
	radiobutton $w.pan.coord.physical -text "Physical" \
	    -variable panZoom(coord) -value physical -selectcolor $menu(selectcolor)
	radiobutton $w.pan.coord.image -text "Image" \
	    -variable panZoom(coord) -value image -selectcolor $menu(selectcolor)

	if {[$current(frame) has wcs equatorial]} {
	    pack $w.pan.coord.fk4 $w.pan.coord.fk5 $w.pan.coord.icrs \
		$w.pan.coord.galactic \
		$w.pan.coord.ecliptic -side top -anchor w -padx 10
	}

	if {[$current(frame) has mosaic] && [$current(frame) has physical]} {
	    pack $w.pan.coord.detector -side top -anchor w -padx 10
	}

	if {[$current(frame) has mosaic]} {
	    pack $w.pan.coord.amplifier -side top -anchor w -padx 10
	}

	if {[$current(frame) has physical]} {
	    pack $w.pan.coord.physical -side top -anchor w -padx 10
	}

	pack $w.pan.coord.image -side top -anchor w -padx 10
	
	pack $w.pan.title $w.pan.value $w.pan.coord -side left -anchor nw -padx 4 -pady 4

	button $w.buttons.ok -text "OK" -default active -command {set panZoom(ok) 1}
	button $w.buttons.apply -text "Apply" -command PanZoomApply
	button $w.buttons.cancel -text "Cancel" -command {set panZoom(ok) 0}
	pack $w.buttons.ok $w.buttons.apply $w.buttons.cancel \
	    -side left -padx 10 -expand true

	bind $w <Return> {set panZoom(ok) 1}
	bind $w <Alt-o> "tkButtonInvoke $w.buttons.ok"
	bind $w <Alt-c> "tkButtonInvoke $w.buttons.cancel"
    }

    DialogCenter $w 
    $w.zoom.value select range 0 end
    DialogWait $w panZoom(ok) $w.zoom.value
    DialogDismiss $w

    if {$panZoom(ok)} {
	PanZoomApply
    }

    # we want to destroy this window

    destroy $w 
}

proc PanZoomApply {} {
    global current
    global panZoom

    if {$current(frame) != ""} {
	set current(zoom) $panZoom(zoom)
	set current(rotate) $panZoom(rotate)

	SetWatchCursor
	$current(frame) zoom to $current(zoom) about \
	    $panZoom(x) $panZoom(y) $panZoom(coord)
	$current(frame) rotate to $current(rotate)
	UpdateLockFrames \
	    "zoom to $current(zoom) about $panZoom(x) $panZoom(y) $panZoom(coord)"
	UpdateLockFrames [list rotate to $current(rotate)]
	UpdateGraphXAxis
	UnsetWatchCursor
    }
}

proc UpdatePanZoomMenu {} {
    global current
    global wcs
    global buttons

    if {$current(frame) != ""} {
	.menuBar entryconfig Zoom -state normal
	$buttons(name).major.zoom configure -state normal

	if {[$current(frame) get fits count] > 0} {
	    .menuBar.zoom entryconfig "Zoom to Fit Frame" -state normal
	} else {
	    .menuBar.zoom entryconfig "Zoom to Fit Frame" -state disabled
	}

	set current(zoom) [$current(frame) get zoom]
	set current(rotate) [$current(frame) get rotate]
	set current(orient) [$current(frame) get orient]
	set wcs(align) [$current(frame) get wcs align]
    } else {
	.menuBar entryconfig Zoom -state disabled
	$buttons(name).major.zoom configure -state disabled

	set current(zoom) 1
	set current(rotate) 0
	set current(orient) none
	set wcs(align) 0
    }
}

