#include <string.h>

#include <strstream.h>
#include <iomanip.h>

#include "card.h"

FitsCard::FitsCard()
{
  card_ = new char[FT_CARDLEN];
  managed = 1;
}

FitsCard::FitsCard(char* c)
{
  card_ = c;
  managed = 0;
}

FitsCard::FitsCard(const FitsCard& a)
{
  card_ = new char[FT_CARDLEN];
  managed = 1;
  memcpy(card_, a.card_, FT_CARDLEN);
}

FitsCard::~FitsCard()
{
  if (managed)
    delete [] card_;
}

FitsCard& FitsCard::operator=(const FitsCard& a)
{
  memcpy(card_, a.card_, FT_CARDLEN);
}

FitsCard& FitsCard::clear()
{
  memset(card_, ' ', FT_CARDLEN);
  return *this;
}

FitsCard& FitsCard::setKey(const char* name)
{
  if (name) {
    memset(card_, ' ', 8);
    ostrstream str(card_, 8);
    str << name;
  }
  return *this;
}

FitsCard& FitsCard::setLogical(const char* name, int value, const char* comment)
{
  setKey(name);

  memset(card_+8, ' ', FT_CARDLEN-8);
  ostrstream str(card_+8, FT_CARDLEN-8);
  str << "=                    " << (value ? 'T' : 'F');
  if (comment)
    str << " / " << comment;
}

FitsCard& FitsCard::setInteger(const char* name, int value, const char* comment)
{
  setKey(name);

  memset(card_+8, ' ', FT_CARDLEN-8);
  ostrstream str(card_+8, FT_CARDLEN-8);
  str << "= " << setw(20) << value;
  if (comment)
    str << " / " << comment;
}

FitsCard& FitsCard::setReal(const char* name, double value, 
			    int prec, const char* comment)
{
  setKey(name);

  memset(card_+8, ' ', FT_CARDLEN-8);
  ostrstream str(card_+8, FT_CARDLEN-8);
  str << "= " << setw(20) << setprecision(prec) << value;
  if (comment)
    str << " / " << comment;
}

FitsCard& FitsCard::setComplex(const char* name, double real, double img, 
			int prec, const char* comment)
{
  setKey(name);

  memset(card_+8, ' ', FT_CARDLEN-8);
  ostrstream str(card_+8, FT_CARDLEN-8);
  str << "= " << setw(20) << setprecision(prec) << real << img;
  if (comment)
    str << " / " << comment;
}

FitsCard& FitsCard::setString(const char* name, const char* value, const char* comment)
{
  setKey(name);

  memset(card_+8, ' ', FT_CARDLEN-8);
  ostrstream str(card_+8, FT_CARDLEN-8);
  str << "= '" << value << '\'';
  if (comment)
    str << " / " << comment;
}

FitsCard& FitsCard::setComment(const char* name, const char* value)
{
  setKey(name);

  memset(card_+8, ' ', FT_CARDLEN-8);
  ostrstream str(card_+8, FT_CARDLEN-8);
  str << "= " << value;
}

int FitsCard::getLogical()
{
  for (int i=10; i<80; i++)
    if (card_[i] != ' ')
      return (card_[i] == 'T' || card_[i] == 't');

  return 0;
}

int FitsCard::getInteger()
{
  istrstream str(card_+10, FT_CARDLEN-10);
  int r;
  str >> r;
  return r;
}

double FitsCard::getReal()
{
  istrstream str(card_+10, FT_CARDLEN-10);
  double r;
  str >> r;
  return r;
}

void FitsCard::getComplex(double* r, double* i)
{
  istrstream str(card_+10, FT_CARDLEN-10);
  str >> *r >> *i;
}

char* FitsCard::getString()
{
  char value[FT_CARDLEN];
  value[0] = '\0';

  if (card_[10] == '\'') {
    int i,j;
    for (i=11, j=0; i < FT_CARDLEN; i++, j++ ) {
      if (card_[i] == '\'')
	if (i+1 == FT_CARDLEN || card_[i+1] != '\'') {
	  break;
	}
	else
	  i++;

      value[j] = card_[i];
    }
    for (; j && value[j-1]==' '; j--);
    value[j] = '\0';
  }
  else {
    int i,j;
    for (i=10; i<FT_CARDLEN; i++)
      if (card_[i] != ' ')
	break;

    for (j=0; i<FT_CARDLEN; i++, j++) {
      if (card_[i] == '/')
	break;

      value[j] = card_[i];
    }
    value[j] = '\0';
  }

  char* cpy = new char[strlen(value)+1];
  strcpy(cpy, value);

  return cpy;
}

char* FitsCard::getComment()
{
  char* cpy = new char[FT_CARDLEN-7];
  memcpy(cpy, card_+8, FT_CARDLEN-8);
  cpy[FT_CARDLEN-8] = '\0';

  return cpy;
}

