#include <stdlib.h>
#include <string.h>
#include <math.h>

#include "colorscale.h"
#include "util.h"

ColorScale::ColorScale(int s)
{
  size_ = s;
  psColors_ = new unsigned char[s*3];
  psIndex_ = new unsigned short[s];

  memset(psColors_, '0', size_*3);
  memset(psIndex_, '0', size_*sizeof(unsigned short));
}

ColorScale::~ColorScale()
{
  if (psColors_)
    delete psColors_;

  if (psIndex_)
    delete psIndex_;
}

LinearScale::LinearScale(int s, unsigned short* indexCells, unsigned char* colorCells,
			 int count) : ColorScale(s)
{
  for (int i=0; i<s; i++) {
    int l = (int)((double)i/s * count);
    psIndex_[i] = indexCells[l];
    memcpy(psColors_+i*3, colorCells+l*3,3);
  }
}

/* no longer used
 * this algorithm has some series problems.
LnScale::LnScale(int s, unsigned index* indexCells, unsigned char* colorCells,
		 int count) : ColorScale(s)
{
  // y = log((x/xSize*(e**10-1)+1)/10 * ySize
  // This algorthim has the following properties:
  // 1. y range log curve from 0 to 1
  // 2. x range independent

  double exp1 = exp(10.)-1;
  for (int i=0; i<s; i++) {
    int l = (int)(log(((double)i/s*exp1)+1)/10 * count);
    psIndex_[i] = indexCells[l];
    memcpy(psColors_+i*3, colorCells+l*3,3);
  }
}
*/
LogScale::LogScale(int s, unsigned short* indexCells, unsigned char* colorCells,
		   int count) : ColorScale(s)
{
  for (int i=0; i<s; i++) {
    int l = (int)((log10(i+1)/log10(s))*count); // will spill over at the top
    if (l >= count)
      l = count-1;
    psIndex_[i] = indexCells[l];
    memcpy(psColors_+i*3, colorCells+l*3,3);
  }
}

SquaredScale::SquaredScale(int s, unsigned short* indexCells, unsigned char* colorCells,
			   int count) : ColorScale(s)
{
  for(int i=0; i<s; i++) {
    double a = (double)i/s;
    int l = (int)(a*a * count);
    psIndex_[i] = indexCells[l];
    memcpy(psColors_+i*3, colorCells+l*3,3);
  }
}

SqrtScale::SqrtScale(int s, unsigned short* indexCells, unsigned char* colorCells,
		     int count) : ColorScale(s)
{
  for(int i=0; i<s; i++) {
    int l = (int)(sqrt((double)i/s) * count);
    psIndex_[i] = indexCells[l];
    memcpy(psColors_+i*3, colorCells+l*3,3);
  }
}

