#include <stdlib.h>
#include <string.h>
#include <math.h>
#include <float.h>

#include <iostream.h>
#include <strstream.h>
#include <fstream.h>
#include <iomanip.h>

#include <Xlib.h>
#include <Xutil.h>

#include "frame.h"
#include "util.h"
#include "fitsimage.h"
#include "nan.h"

#include "contour.h"

#include "hist.h"

#define BLT 1

#if BLT == 1
#include "blt.h"
#endif

// Parser Stuff

#undef yyFlexLexer
#define yyFlexLexer frFlexLexer
#include <FlexLexer.h>

frFlexLexer* frlexx = NULL;    // used by frlex
Frame* fr = NULL;           // used by frerror
extern int frparse(void*);

int frlex()
{
  return (frlexx ? frlexx->yylex() : 0);
}

void frerror(const char* m)
{
  if (fr) {
    fr->error(m);
    const char* cmd = frlexx ? frlexx->YYText() : (const char*)NULL;
    if (cmd && cmd[0] != '\n') {
      fr->error(": ");
      fr->error(cmd);
    }
  }
}

// Frame Member Functions

Frame::Frame(Tcl_Interp* i, Tk_Canvas c, Tk_Item* item) : Widget(i, c, item)
{
  name = NULL;
  fits = NULL;
  currentFits = NULL;
  fitsCount = 0;
  basePixmap = 0;
  baseXImage = NULL;

  needsUpdate = NOUPDATE;

  zoom = 1;
  rotation = 0;
  orientation = NORMAL;

  wcsAlignment = 0;
  wcsZoom = 1;
  wcsRotation = 0;
  wcsOrientation = NORMAL;
  
  panPM = 0;
  panGCXOR = NULL;

  rotateRotation = 0;
  rotateSrcXM = NULL;
  rotateDestXM = NULL;
  rotatePM = 0;
  rotateGCXOR = NULL;

  cmapID = 1;
  bias = 0.5;
  contrast = 1.0;
  invert = 0;

  colorCount = 0;

  colorScale = NULL;
  colorScaleType = LINEARSCALE;
  colorCells = NULL;
  indexCells = NULL;

  clipScope = LOCAL;
  clipMode = MINMAX;
  mmMode = SAMPLE;

  min = low = uLow = 1;
  max = high = uHigh = 100;
  scanOnLoad = 0;

  zContrast = .25;
  zSampleSize = 600;
  zSampleLine = 120;

  mmIncr = 10;

  histBins = 64;

  useCrosshair = 0;
  crosshairGC = NULL;
  crosshairGCXOR = NULL;

  highliteGC = NULL;

  useSquarePixels = 0;
  aspect = (DisplayHeight(display,screenNumber) / 
	    DisplayHeightMM(display,screenNumber)) /
           (DisplayWidth(display,screenNumber) / 
	    DisplayWidthMM(display,screenNumber));

  markerGC = NULL;
  markerGCXOR = NULL;
  markerGCXORPoint = NULL;
  markerGCXORRuler = NULL;
  editMarker = NULL;
  rotateMarker = NULL;

  pannerPixmap = 0;
  pannerXImage = NULL;
  pannerWidth = 0;
  pannerHeight = 0;
  pannerName[0] = '\0';
  usePanner = 0;

  magnifierPixmap = 0;
  magnifierXImage = NULL;
  magnifierWidth = 0;
  magnifierHeight = 0;
  magnifierName[0] = '\0';
  magnifierMode = ACCURATE;
  magnifierZoom = 4;
  useMagnifier = 0;
  useMagnifierGraphics = 1;
  useMagnifierCursor = 1;
  magnifierGC = NULL;

  useHighlite = 0;

  contour = NULL;
  contourGC = NULL;

  psLevel = 2;
  psResolution = 300;
  psInterpolate = 0;
  psColorSpace = RGB;

  blockMode = MANUALBLOCK;
  blockFunction = FitsHist::AVERAGE;
  blockFactor = 1;
  blockBufferSize = 1024;

  iisfoo = 0;
}

Frame::~Frame()
{
  if (name)
    delete name;

  unloadFits();

  if (basePixmap)
    Tk_FreePixmap(display, basePixmap);

  if (baseXImage)
    XDestroyImage(baseXImage);

  if (panPM)
    Tk_FreePixmap(display, panPM);

  if (crosshairGC)
    XFreeGC(display, crosshairGC);

  if (crosshairGCXOR)
    XFreeGC(display, crosshairGCXOR);

  if (highliteGC)
    XFreeGC(display, highliteGC);

  if (markerGC)
    XFreeGC(display, markerGC);

  if (markerGCXOR)
    XFreeGC(display, markerGCXOR);

  if (markerGCXORPoint)
    XFreeGC(display, markerGCXORPoint);

  if (markerGCXORRuler)
    XFreeGC(display, markerGCXORRuler);

  if (panGCXOR)
    XFreeGC(display, panGCXOR);

  if (rotateGCXOR)
    XFreeGC(display, rotateGCXOR);

  if (pannerPixmap)
    Tk_FreePixmap(display, pannerPixmap);

  if (pannerXImage)
    XDestroyImage(pannerXImage);

  if (magnifierPixmap)
    Tk_FreePixmap(display, magnifierPixmap);

  if (magnifierXImage)
    XDestroyImage(magnifierXImage);

  if (magnifierGC)
    XFreeGC(display, magnifierGC);

  if (colorScale)
    delete colorScale;

  if (colorCells)
    delete [] colorCells;

  if (indexCells)
    delete [] indexCells;

  if (contour)
    delete contour;

  if (contourGC)
    XFreeGC(display, contourGC);
}

int Frame::parse(istrstream& istr)
{
  result = TCL_OK;
  frlexx = new frFlexLexer(&istr);
  fr = this;
  frparse(this);

  delete frlexx;
  frlexx = NULL;
  return result;
}

void Frame::reset()
{
  zoom = 1;
  rotation = 0;
  orientation = NORMAL;
  orientationMatrix.identity();

  wcsAlignment = 0;
  wcsZoom = 1;
  wcsRotation = 0;
  wcsOrientation = NORMAL;
  wcsOrientationMatrix.identity();
  
  cmapID = 1;
  bias = 0.5;
  contrast = 1.0;
  invert = 0;

  centerImage();

  unselectMarkers();

  crosshair = cursor;

  update(MATRIX);
}

// Rendering Routines

void Frame::invalidPixmap()
{
  Widget::invalidPixmap();

  if (basePixmap)
    Tk_FreePixmap(display, basePixmap);
  basePixmap = 0;

  if (baseXImage)
    XDestroyImage(baseXImage);
  baseXImage = NULL;

  needsUpdate = MATRIX;
}

void Frame::update(UpdateType flag)
{
  // Be careful, someone may have already set the flag at a lower level
  // therefor, only change the flag if we need more to be done

  if (flag < needsUpdate)
    needsUpdate = flag;
  redraw();
}

void Frame::update(UpdateType flag, BBox bb)
{
  // bb is in canvas coords

  if (flag < needsUpdate)
    needsUpdate = flag;
  redraw(bb);
}

void Frame::updateNow(UpdateType flag)
{
  if (flag < needsUpdate)
    needsUpdate = flag;
  redrawNow();
}

void Frame::updateNow(UpdateType flag, BBox bb)
{
  // bb is in canvas coords

  if (flag < needsUpdate)
    needsUpdate = flag;
  redrawNow(bb);
}

int Frame::updatePixmap(const BBox& bb)
{

  // Note: lack of breaks-- on purpose. 
  //If Matrices are update, both Base and Pixmap
  // also need to be updated. Same for Base-- ie, pixmap is also updated.
  // bb is in canvas coords

  switch (needsUpdate) {
  case MATRIX:
    updateMatrices();
    updateMarkers();
  case BASE:
    updateBase();
  case PIXMAP:
    updatePM(bb);
    break;
  }

  needsUpdate = NOUPDATE;
  return TCL_OK;
}

void Frame::updateBase()
{
  //  XSync(display,0);
  if (DebugPerf)
    cerr << "updateBase...";

  int& width = options->width;
  int& height = options->height;

  if (!basePixmap) {
    if (!(basePixmap = Tk_GetPixmap(display, Tk_WindowId(tkwin), 
				    width, height, depth))){
      cerr << "Frame Internal Error: Unable to Create Pixmap" << endl;
      exit(1);
    }

    // Geometry has changed, redefine our marker GCs including clip regions

    updateGCs();
  }

  if (!baseXImage) {
    baseXImage = XGetImage(display, basePixmap, 0, 0, width, height, 
			   AllPlanes, ZPixmap);
    if (!baseXImage) {
      cerr << "Frame Internal Error: Unable to Create XImage" << endl;
      exit(1);
    }

    // we have a race condition. Some Truecolor ColorScales need to know the 
    // bytes per pixel, RGB masks, and byte order, from XImage struct.
    // yet, we receive colormap commands well before we render to the 
    // screen and have a valid XImage. 
    // So, we put off creating a colorscale until we are ready to render.

    if (!colorScale)
      updateColorScale();
  }

  if (fits)
    ximageToPixmap(basePixmap, baseXImage, width, height, 
		   &FitsImage::getWidgetToData);
  else
    clearPixmap(basePixmap, baseXImage);
  
  renderOverlay();   // render any overlay graphics into pixmap

  updatePanner();

  if (DebugPerf)
    cerr << "end" << endl;
}

void Frame::updatePM(const BBox& bbox)
{
  if (DebugPerf)
    cerr << "updatePM...";

  int& width = options->width;
  int& height = options->height;

  if (!pixmap)
    if (!(pixmap = Tk_GetPixmap(display, Tk_WindowId(tkwin), 
				width, height, depth))) {
      cerr << "Frame Internal Error: Unable to Create Pixmap" << endl;
      exit(1);
    }

  if (!bbox.isEmpty()) {
    BBox bb = bbox * canvasToWidget;
    Vector size = bb.size();
    XCopyArea(display, basePixmap, pixmap, gc, 
	      (int)bb.ll[0], (int)bb.ll[1], (int)size[0], (int)size[1],
	      (int)bb.ll[0], (int)bb.ll[1]);
  }

  renderMarkers(bbox); // render the markers into pixmap
  renderCrosshair(); // render crosshair into pixmap
  renderHighLite();  // render a hightlite bbox into pixmap

  if (DebugPerf)
    cerr << "end" << endl;
}

Vector Frame::getCenter()
{
  return fits ? fits->getCenter() : Vector();
}

void Frame::updateMatrices()
{
  if (DebugPerf)
    cerr << "updateMatrices...";

  // These are the basic tranformation matrices

  Vector center = getCenter();

  refToUser = 
    Translate(-center) *                // translate to center image
    FlipY() *                           // flip y axis for X Windows
    Translate(center);                  // translate back from center
  userToRef = refToUser.invert();

  userToWidget =
    Translate(-cursor) *                // translate to cursor position
    Rotate(rotation) *                  // rotate about cursor position
    Rotate(wcsRotation) *               // rotate about center position
    orientationMatrix *                 // flip x/y axis about cursor position
    wcsOrientationMatrix *              // flip x/y axis about center
    Scale(zoom) *                       // scale about cursor position
    Scale(wcsZoom) *                    // scale about wcs scale
    aspectRatioMatrix *                 // scale about aspect ratio
    Translate(cursor) *                 // translate back from cursor

    Translate(-cursor) *                // translate to cursor positon

 // translate to center of widget. needs to be integer

    Translate((int)(options->width/2.), (int)(options->height/2.));

  widgetToUser = userToWidget.invert();

  widgetToCanvas = Translate(originX, originY); // translate by widget location
  canvasToWidget = widgetToCanvas.invert();

  short x, y;
  Tk_CanvasWindowCoords(canvas, 0, 0, &x, &y);
  canvasToWindow = Translate(x, y);
  windowToCanvas = canvasToWindow.invert();

  // These are derived Transformation Matrices

  refToWidget = refToUser * userToWidget;
  widgetToRef = refToWidget.invert();

  refToCanvas = refToUser * userToWidget * widgetToCanvas;
  canvasToRef = refToCanvas.invert();

  refToWindow = refToUser * userToWidget * widgetToCanvas * canvasToWindow;
  windowToRef = refToWindow.invert();

  userToCanvas = userToWidget * widgetToCanvas;
  canvasToUser = userToCanvas.invert();

  widgetToWindow = widgetToCanvas * canvasToWindow;
  windowToWidget = widgetToWindow.invert();

    // calculate panner matrices

  userToPanner =
    Translate(-center) *
    Rotate(rotation) *                  // rotate about cursor position
    Rotate(wcsRotation) *               // rotate about center position
    orientationMatrix *                 // flip x/y axis about cursor position
    wcsOrientationMatrix *              // flip x/y axis about center
    Scale(calcPannerZoom()) *
    aspectRatioMatrix *                 // scale about aspect ratio
    Translate(pannerWidth/2., pannerHeight/2.);
  pannerToUser = userToPanner.invert();

  refToPanner = refToUser * userToPanner;
  pannerToRef = refToPanner.invert();

  updateMagnifierMatrix();

  // update all image matrices

  FitsImage* ptr = fits;
  while (ptr) {
    ptr->updateMatrices(refToUser, userToWidget, widgetToCanvas, userToPanner);
    ptr = ptr->next();
  }

  if (DebugPerf)
    cerr << "end" << endl;
}

void Frame::ximageToPixmap(Pixmap pmap, XImage* xmap, int width, int height, 
			   Matrix& (FitsImage::*getMatrix)())
{
  // generate new image

  switch (fitsCount) {
  case 0:
    return;
  case 1:
    fillXImage(xmap, getMatrix);
    break;
  default:;
    fillXImageMosaic(xmap, getMatrix);
    break;
  }

  // update

  XPutImage(display, pmap, gc, xmap, 0, 0, 0, 0, width, height);
}

void Frame::renderOverlay()
{
  if (contour)
    contour->render();
}

void Frame::renderCrosshair()
{
  Vector r = crosshair * refToWidget;

  if (useCrosshair) {
    XDrawLine(display, pixmap, crosshairGC, 
	      (int)r[0], 1, (int)r[0], options->height);
    XDrawLine(display, pixmap, crosshairGC,
	      1, (int)r[1], options->width, (int)r[1]);
  }
}

void Frame::drawCrosshair()
{
  if (visible) {
    Vector r = crosshair * refToWidget;

    // Vertical line

    BBox vv = BBox(Vector(r[0], 1), Vector(r[0], options->height)) * 
      widgetToWindow;
    XDrawLine(display, Tk_WindowId(tkwin), crosshairGCXOR, 
	      (int)vv.ll[0], (int)vv.ll[1], (int)vv.ur[0], (int)vv.ur[1]);

    // Horizontal line

    BBox hh = BBox(Vector(1, r[1]), Vector(options->width, r[1])) * 
      widgetToWindow;
    XDrawLine(display, Tk_WindowId(tkwin), crosshairGCXOR, 
	      (int)hh.ll[0], (int)hh.ll[1], (int)hh.ur[0], (int)hh.ur[1]);
  }
}

void Frame::eraseCrosshair()
{
  if (visible) {
    // This routine schedules an update for only the area written to previously
    // by drawCursor.

    Vector r = crosshair * refToWidget;

    // Vertical line

    BBox vv = BBox(Vector(r[0], 1), Vector(r[0], options->height)) * 
      widgetToCanvas;
    redrawNow(vv.expand(1));

    // Horizontal line

    BBox hh = BBox(Vector(1, r[1]), Vector(options->width, r[1])) * 
      widgetToCanvas;
    redrawNow(hh.expand(1));
  }
}

void Frame::renderHighLite()
{
  if (useHighlite)
    XDrawRectangle(display, pixmap, highliteGC, 1, 1, 
		   options->width-2, options->height-2);
}

// Other Private Member Functions

void Frame::wcsAlign()
{
  if (currentFits && currentFits->hasWCS() && 
      currentFits->hasWCSEqu() && wcsAlignment) {
    wcsZoom = 1;
    wcsRotation = currentFits->getWCSRotation();
    wcsOrientation = currentFits->getWCSOrientation();
  }
  else {
    wcsZoom = 1;
    wcsRotation = 0;
    wcsOrientation = NORMAL;
  }

  switch (wcsOrientation) {
  case NORMAL:
    wcsOrientationMatrix.identity();
    break;
  case X:
    wcsOrientationMatrix = FlipX();
    break;
  case Y:
    wcsOrientationMatrix = FlipY();
    break;
  case XY:
    wcsOrientationMatrix = FlipXY();
    break;
  }

}

void Frame::graphCut(char* xName, char* yName, int axis, const Vector& rr)
{
#if BLT == 1
  // Vector r is in Widget coords
  Vector r = rr;

  int size;
  if (axis == Frame::X)
    size = options->width;
  else
    size = options->height;

  long length = (size+1) * 2;
  double* x = (double*)malloc(sizeof(double)*length);
  double* y = (double*)malloc(sizeof(double)*length);

  switch (fitsCount) {
  case 0:
    for (int i=0; i<size; i++) {
      x[i*2] = i;
      x[i*2+1] = i;
      y[i*2] = 0;
      y[i*2+1] = 0;
    }
    break;
  case 1:
    fillArray(x, y, size, r, axis);
    break;
  default:;
    fillArrayMosaic(x, y, size, r, axis);
    break;
  }

  Blt_Vector* xx;
  if (Blt_GetVector(interp, xName, &xx) != TCL_OK)
    goto error;

  if (Blt_ResetVector(xx, x, length, length*sizeof(double), TCL_DYNAMIC) != 
      TCL_OK)
    goto error;

  Blt_Vector* yy;
  if (Blt_GetVector(interp, yName, &yy) != TCL_OK)
    goto error;

  if (Blt_ResetVector(yy, y, length, length*sizeof(double), TCL_DYNAMIC) != 
      TCL_OK)
    goto error;

  return;

 error:
    result = TCL_ERROR;
    return;
#endif
}

void Frame::fillArray(double* x, double* y, int size, Vector& r, int axis)
{
  Matrix m = fits->getWidgetToData();
  register int width = fits->width();
  register int height = fits->height();
  double prev = low;

  for (int i=0; i<=size; i++) {
    double v = low;

    Vector img;
    if (axis == Frame::X)
      img = Vector(1+i,r[1]) * m;
    else
      img = Vector(r[0],1+i) * m;

    if (img[0]>=0 && img[0]<width && img[1]>=0 && img[1]<height) {
      register double value = fits->getValueDouble(img);
      if (!isnand(value))
	v = value;
    }

    x[2*i] = i;
    x[2*i +1] = i;
    
    y[2*i] = prev;
    y[2*i +1] = v;
    prev = v;
  }
}

void Frame::fillArrayMosaic(double* x, double* y, int size, 
			    Vector& r, int axis)
{
  Matrix m;
  register int width;
  register int height;

  double prev = low;
  int found = 0;
  FitsImage* current = fits;
  FitsImage* ptr;

  for (int i=0; i<=size; i++) {
    double v = low;

    // find y first

  tryagain:
    if (!found) {
      ptr = current;
      current = current->next();

      m = ptr->getWidgetToData();
      width = ptr->width();
      height = ptr->height();
    }

    Vector img;
    if (axis == Frame::X)
      img = Vector(1+i,r[1]) * m;
    else
      img = Vector(r[0],1+i) * m;

    if (img[0]>=0 && img[0]<width && img[1]>=0 && img[1]<height) {
      register double value = ptr->getValueDouble(img);
      if (!isnand(value))
	v = value;
      found = 1;
    }
    else {
      found = 0;
      if (current)
	goto tryagain;
      else
	current = fits;
    }

    x[2*i] = i;
    x[2*i +1] = i;
    
    y[2*i] = prev;
    y[2*i +1] = v;
    prev = v;
  }
}

void Frame::histogram(char* xName, char* yName)
{
#if BLT == 1
  updateMinMax();

  double* x = (double*)calloc(histBins, sizeof(double));
  double* y = (double*)calloc(histBins, sizeof(double));

  for (int i=0; i<histBins; i++)
    x[i] = (double)i/(histBins-1)*(max-min) + min;
    
  if (max > min) {
    FitsImage* ptr = fits;
    while (ptr) {
      ptr->bin(y,histBins,min,max);
      ptr = ptr->next();
    }
  }

  Blt_Vector* xx;
  if (Blt_GetVector(interp, xName, &xx) != TCL_OK)
    goto error;

  if (Blt_ResetVector(xx, x, histBins, histBins*sizeof(double), TCL_DYNAMIC) 
      != TCL_OK)
    goto error;

  Blt_Vector* yy;
  if (Blt_GetVector(interp, yName, &yy) != TCL_OK) 
    goto error;

  if (Blt_ResetVector(yy, y, histBins, histBins*sizeof(double), TCL_DYNAMIC) 
      != TCL_OK)
    goto error;

  return;

 error:
    result = TCL_ERROR;
    return;
#endif
}

void Frame::updateMinMax()
{
  // find over-all min max
  // used by histogram

  min = DBL_MAX;
  max = DBL_MIN;
  FitsImage* ptr = fits;
  while (ptr) {
    ptr->updateMinMax(mmMode, mmIncr);
    if (min > ptr->getMinDouble())
      min = ptr->getMinDouble();

    if (max <= ptr->getMaxDouble())
      max = ptr->getMaxDouble();

    // this is a kluge to fix a problem the compiler
    if (max == DBL_MIN && ptr->getMaxDouble() == 0)
      max = ptr->getMaxDouble();

    ptr = ptr->next();
  }
}

void Frame::updateClip()
{
  if (fits) {
    switch (clipScope) {
    case GLOBAL:
      {
	FitsImage* ptr;

	// calculate clip values for each fits

	ptr = fits;
	while (ptr) {
	  ptr->updateClip(clipMode, mmMode, zContrast, zSampleSize, 
			  zSampleLine, mmIncr, uLow, uHigh, scanOnLoad);
	  ptr = ptr->next();
	}

	updateLowHigh();

	ptr = fits;
	while (ptr) {
	  ptr->setClip(low, high);
	  ptr = ptr->next();
	}
      }
      break;
    case LOCAL:
      {
	FitsImage* ptr = fits;
	while (ptr) {
	  ptr->updateClip(clipMode, mmMode, zContrast, zSampleSize, 
			  zSampleLine, mmIncr, uLow, uHigh, scanOnLoad);
	  ptr = ptr->next();
	}

	updateLowHigh();
      }
      break;
    }

    update(BASE);
  }
  else {
    if (clipMode != USERCLIP) {
      low = 1;
      high = 100;
    }
    else {
      low = uLow;
      high = uHigh;
    }
  }
}

void Frame::updateLowHigh()
{
  // find over-all low high

  low = DBL_MAX;
  high = DBL_MIN;
  FitsImage* ptr = fits;
  while (ptr) {
    if (low > ptr->getLowDouble())
      low = ptr->getLowDouble();

    if (high <= ptr->getHighDouble())
      high = ptr->getHighDouble();

    // this is a kluge to fix a problem the compiler
    if (high == DBL_MIN && ptr->getHighDouble() == 0)
      high = ptr->getHighDouble();

    ptr = ptr->next();
  }
}

void Frame::unloadFits()
{
  if (contour)
    delete contour;
  contour = NULL;

  while (fits) {
    FitsImage* tmp = fits->next();
    delete fits;
    fits = tmp;
  }
  fits = NULL;
  currentFits = NULL;
  fitsCount = 0;
}

double Frame::calcZoom(int width, int height, Vector ur)
{
  // we need to calculate the width and height of the rotated image
  // so we can derived a zoom factor to shrink it to fit the requested size

  Vector ll(0,0);
  Vector ul(ll[0],ur[1]);
  Matrix m = Translate(-ur/2.) * Rotate(wcsRotation) * Rotate(rotation);

  Vector cc = ll * m;
  Vector dd = ul * m;

  float z1 = fabs(cc[0]) > fabs(cc[1]) ? fabs(cc[0]) : fabs(cc[1]);
  float z2 = fabs(dd[0]) > fabs(dd[1]) ? fabs(dd[0]) : fabs(dd[1]);

  return ((width<height ? width:height) / ((z1>z2 ? z1:z2)*2));
}

// Color support

void Frame::updateColorCells(unsigned short* index, 
			     unsigned char* cells, int cnt)
{
  // the colorbar widget will pass us a pointer to the indexCells

  colorCount = cnt;
  if (indexCells)
    delete [] indexCells;
  indexCells = new unsigned short[cnt];
  if (!indexCells) {
    cerr << "Frame Internal Error: Unable to Alloc indexCells" << endl;
    exit(1);
  }
  memcpy(indexCells, index, cnt*sizeof(unsigned short));

  // copy the rgb vales to the colorCells array (for postscript printing)

  if (colorCells)
    delete [] colorCells;
  colorCells = new unsigned char[cnt*3];
  if (!colorCells) {
    cerr << "Frame Internal Error: Unable to Alloc colorCells" << endl;
    exit(1);
  }
  memcpy(colorCells, cells, cnt*3);
}

// Panner Support

double Frame::calcPannerZoom()
{
  if (pannerPixmap && fits) {
    calcZoom(pannerWidth,pannerHeight,fits->getSize());
  }
}

void Frame::updatePanner()
{
  if (usePanner)
    if (fits)
      ximageToPixmap(pannerPixmap, pannerXImage, pannerWidth, pannerHeight,
		     &FitsImage::getPannerToData);
    else
      XFillRectangle(display, pannerPixmap, gc, 0, 0, 
		     pannerWidth, pannerHeight);
  
  pannerPush();
}

void Frame::pannerPush()
{
  if (usePanner) {
    char buf[512];
    ostrstream str(buf, 512);
    if (fits) {

      // calculate bbox

      Matrix widgetToPanner = widgetToUser * userToPanner;
      BBox bb = BBox(Vector(0,0), Vector(options->width, options->height)) * 
	widgetToPanner;

      // calculate compass vectors

      Rotate orient = Rotate(wcsRotation) * Rotate(rotation) * 
	wcsOrientationMatrix * orientationMatrix;

      str << pannerName << " update " << pannerPixmap << ";";
      str << pannerName << " update bbox " << bb << ";";
      str << pannerName << " update image compass " << orient << ";";

      if (currentFits->hasWCS()) {
	Matrix m = currentFits->getImageToUser() * userToPanner;

	// find origin of wcs-- we must take the ref point,
	// run it thru pix2wcs (which may convert to J2000 if needed)
	// and then run it back thru wcs2pix to find our ref point in either
	// J2000 or B1950

	Vector orval = currentFits->pix2wcs(currentFits->getWCScrpix());
	Vector orpix = currentFits->wcs2pix(orval);
	Vector origin = orpix * m;

	// find normalized north & east vectors

	Vector delta = (currentFits->getWCScdelt() * 10).abs();

	Vector north = 
	  currentFits->wcs2pix(Vector(orval[0],orval[1]+delta[1])) * m;
	Vector ndiff = north - origin;

	Vector east = 
	  currentFits->wcs2pix(Vector(orval[0]+delta[0],orval[1])) * m;
	Vector ediff = east - origin;

	// and update the panner

	str << pannerName << " update wcs compass " << origin
	    << ndiff.normalize() << ediff.normalize() << ends;
      }
      else
	str << pannerName << " update wcs compass invalid" << ends;
    }
    else
      str << pannerName << " clear" << ends;

    Tcl_Eval(interp, buf);
  }
}

// Magnifier Support

void Frame::updateMagnifier()
{
  updateMagnifier(magnifierCursor);
}

void Frame::updateMagnifier(const Vector& v)
{
  // v is in CANVAS coords
  // save it, we may need it later
  magnifierCursor = v;

  if (useMagnifier && magnifierXImage && magnifierGC && fits) {

    // clear XImage data

    memset(magnifierXImage->data, 0, magnifierXImage->bytes_per_line * 
	   magnifierHeight);

    // fill XImage

    switch (magnifierMode) {
    case FAST:
      updateMagnifierReplicate(v);
      break;
    case ACCURATE:
      updateMagnifierZoom(v); // pixmap is updated inside
      break;
    }

    // render cursor
    
    if (useMagnifierCursor)
      renderMagnifierCursor(v);

    // notify the magnifier widget

    char buf[64];
    ostrstream str(buf, 64);
    str << magnifierName << " update " << magnifierPixmap << ends;

    Tcl_Eval(interp, buf);
  }
}

void Frame::updateMagnifierMatrix()
{
  // calculate partial magnifier matrix
  // This is a partial matix, it lacks the Translate at the begining

  userToMagnifier =                     
    Rotate(rotation) *                  // rotate about cursor position
    Rotate(wcsRotation) *               // rotate about center position
    orientationMatrix *                 // flip x/y axis about cursor position
    wcsOrientationMatrix *              // flip x/y axis about center
    Scale(zoom) *                       // scale about cursor position
    Scale(wcsZoom) *                    // scale about wcs scale
    Scale(magnifierZoom) *              // scale
    aspectRatioMatrix *
    Translate(magnifierWidth/2., magnifierHeight/2.);

  userToMagnifierCursor =                     
    Rotate(rotation) *
    Rotate(wcsRotation) *
    Scale(zoom) *
    Scale(wcsZoom) *
    Scale(magnifierZoom) *
    aspectRatioMatrix *
    Translate(magnifierWidth/2., magnifierHeight/2.);

  Vector magCorrection = Vector(.5,-.5) *
    Rotate(rotation) *
    Rotate(wcsRotation) *
    orientationMatrix *
    wcsOrientationMatrix;

  ximageToMagnifier = 
    Translate(magCorrection) * // we need to offset by this correction
    Scale(magnifierZoom) *
    Translate(magnifierWidth/2.,magnifierHeight/2.);
}

void Frame::updateMagnifierZoom(const Vector& v)
{
  // v is in CANVAS coords
  Matrix refToMagnifier = refToUser * Translate(-(v * canvasToUser)) * 
    userToMagnifier;
  Matrix magnifierToRef = refToMagnifier.invert();

  // Update magnifier matrices

  FitsImage* ptr = fits;
  while (ptr) {
    ptr->updateMagnifierMatrices(refToMagnifier);
    ptr = ptr->next();
  }

  // Update ximage

  switch (fitsCount) {
  case 0:
    return;
  case 1:
    fillXImage(magnifierXImage, &FitsImage::getMagnifierToData);
    break;
  default:;
    fillXImageMosaic(magnifierXImage, &FitsImage::getMagnifierToData);
    break;
  }

  // copy XImage to Pixmap

  XPutImage(display, magnifierPixmap, gc, magnifierXImage, 0, 0, 0, 0, 
	    magnifierWidth, magnifierHeight);

  if (useMagnifierGraphics) {

    // render markers
    // markers bounding box is in canvas coords
    // map the magnifier to a bounding box in canvas coords
    Matrix m = magnifierToRef * refToCanvas;
    BBox bb = BBox(v,v);
    bb.bound(Vector(0,0) * m);
    bb.bound(Vector(magnifierWidth,magnifierHeight) * m);
    renderMagnifierMarkers(refToMagnifier, bb);

    // render crosshair
    if (useCrosshair) {
      Vector r = crosshair * refToMagnifier;
      if (r[0]>=0 && r[0]<magnifierWidth)
	XDrawLine(display, magnifierPixmap, crosshairGC, 
		  (int)r[0], 1, (int)r[0], magnifierHeight);

      if (r[1]>=0 && r[1]<magnifierHeight)
	// bump Y by one, I don't know why, it just works
	XDrawLine(display, magnifierPixmap, crosshairGC,
		  1, (int)r[1], magnifierWidth, (int)r[1]);

    }
  }
}

void Frame::renderMagnifierCursor(const Vector& v)
{
    // first, the inner white box

    Vector u = v * canvasToUser;
    Matrix m = Translate(-u) * userToMagnifierCursor;

    Vector c[5];
    c[0] = (u + Vector(-.5,-.5)) * m;
    c[1] = (u + Vector( .5,-.5)) * m;
    c[2] = (u + Vector( .5, .5)) * m;
    c[3] = (u + Vector(-.5, .5)) * m;
    c[4] = c[0];
    
    XPoint pts[5];
    for (int i=0; i<5; i++) {
      Vector z = c[i].round();
      pts[i].x = (int)z[0];
      pts[i].y = (int)z[1];
    }
    XSetForeground(display, magnifierGC, whiteColor->pixel);
    XDrawLines(display, magnifierPixmap, magnifierGC, pts, 5, CoordModeOrigin);

    // now, the outer black box

    Matrix n =
      Translate(-(u * userToMagnifierCursor)) *
      Rotate(-rotation) *
      Rotate(-wcsRotation);
    Matrix o = n.invert();

    Vector d[5];
    for (int i=0; i<5; i++)
      d[i] = c[i] * n;

    d[0]+=Vector(-1,-1);
    d[1]+=Vector( 1,-1);
    d[2]+=Vector( 1, 1);
    d[3]+=Vector(-1, 1);
    d[4]=d[0];

    for (int i=0; i<5; i++) {
      d[i] = d[i] * o;
      Vector z = d[i].round();
      pts[i].x = (int)z[0];
      pts[i].y = (int)z[1];
    }
    XSetForeground(display, magnifierGC, blackColor->pixel);
    XDrawLines(display, magnifierPixmap, magnifierGC, pts, 5, CoordModeOrigin);
}

void Frame::updateGCs()
{
  // widget clip region

  BBox bbWidget = BBox(0, 0, options->width-1, options->height-1);
  Vector sizeWidget = bbWidget.size();

  XRectangle rectWidget[1];
  rectWidget[0].x = (int)bbWidget.ll[0];
  rectWidget[0].y = (int)bbWidget.ll[1];
  rectWidget[0].width = (int)sizeWidget[0];
  rectWidget[0].height = (int)sizeWidget[1];

  // window clip region

  BBox bbWindow = bbWidget * widgetToWindow;
  Vector sizeWindow = bbWindow.size();

  XRectangle rectWindow[1];
  rectWindow[0].x = (int)bbWindow.ll[0];
  rectWindow[0].y = (int)bbWindow.ll[1];
  rectWindow[0].width = (int)sizeWindow[0];
  rectWindow[0].height = (int)sizeWindow[1];

  // crosshairGC

  if (!crosshairGC) {
    crosshairGC = XCreateGC(display, Tk_WindowId(tkwin), 0, NULL);
    XSetForeground(display, crosshairGC, greenColor->pixel);
    XSetLineAttributes(display, crosshairGC, 1, LineSolid, CapButt, JoinMiter);
  }
  XSetClipRectangles(display, crosshairGC, 0, 0, rectWidget, 1, Unsorted);

  // crosshairGCXOR

  if (!crosshairGCXOR) {
    crosshairGCXOR = XCreateGC(display, Tk_WindowId(tkwin), 0, NULL);
    XSetForeground(display, crosshairGCXOR, greenColor->pixel);
    XSetLineAttributes(display, crosshairGCXOR, 1, LineSolid, 
		       CapButt, JoinMiter);
  }
  XSetClipRectangles(display, crosshairGCXOR, 0, 0, rectWindow, 1, Unsorted);

  if (!highliteGC) {
    highliteGC = XCreateGC(display, Tk_WindowId(tkwin), 0, NULL);
    XSetForeground(display, highliteGC, blueColor->pixel);
    XSetLineAttributes(display, highliteGC, 2, LineSolid, CapButt, JoinMiter);
  }
  XSetClipRectangles(display, highliteGC, 0, 0, rectWidget, 1, Unsorted);

  // contourGC

  if (!contourGC) {
    contourGC = XCreateGC(display, Tk_WindowId(tkwin), 0, NULL);
    XSetLineAttributes(display, contourGC, 1, LineSolid, CapButt, JoinMiter);
  }
  XSetClipRectangles(display, contourGC, 0, 0, rectWidget, 1, Unsorted);

  // markerGC

  if (!markerGC)
    markerGC = XCreateGC(display, Tk_WindowId(tkwin), 0, NULL);
  XSetClipRectangles(display, markerGC, 0, 0, rectWidget, 1, Unsorted);

  if (!markerGCXOR) {
    markerGCXOR = XCreateGC(display, Tk_WindowId(tkwin), 0, NULL);
    XSetForeground(display, markerGCXOR, getColor("white"));
    char dlist[] = {8,3};
    XSetLineAttributes(display, markerGCXOR, 1, LineOnOffDash, 
		       CapButt, JoinMiter);
    XSetDashes(display, markerGCXOR, 0, dlist, 2);
  }
  XSetClipRectangles(display, markerGCXOR, 0, 0, rectWindow, 1, Unsorted);

  if (!markerGCXORPoint) {
    markerGCXORPoint = XCreateGC(display, Tk_WindowId(tkwin), 0, NULL);
    XSetForeground(display, markerGCXORPoint, getColor("white"));
  }
  XSetClipRectangles(display, markerGCXORPoint, 0, 0, rectWindow, 1, Unsorted);

  if (!markerGCXORRuler)
    markerGCXORRuler = XCreateGC(display, Tk_WindowId(tkwin), 0, NULL);
  XSetClipRectangles(display, markerGCXORRuler, 0, 0, rectWindow, 1, Unsorted);

  // panGCXOR

  if (!panGCXOR)
    panGCXOR = XCreateGC(display, Tk_WindowId(tkwin), 0, NULL);
  XSetClipRectangles(display, panGCXOR, 0, 0, rectWindow, 1, Unsorted);

  // rotateGCXOR

  if (!rotateGCXOR)
    rotateGCXOR = XCreateGC(display, Tk_WindowId(tkwin), 0, NULL);
  XSetClipRectangles(display, rotateGCXOR, 0, 0, rectWindow, 1, Unsorted);
}
