#include <iostream.h>
#include <strstream.h>
#include <iomanip.h>

#include "../../tk8.0.4/generic/tkCanvas.h"

#include "ps.h"
#include "frame.h"
#include "fitsimage.h"
#include "contour.h"
#include "util.h"
#include "nan.h"

// this is copied from tkCanvPs.c

typedef struct TkPostscriptInfo {
  int x, y, width, height;	/* Area to print, in canvas pixel
				 * coordinates. */
  int x2, y2;			/* x+width and y+height. */
  char *pageXString;		/* String value of "-pagex" option or NULL. */
  char *pageYString;		/* String value of "-pagey" option or NULL. */
  double pageX, pageY;  	/* Postscript coordinates (in points)
			         * corresponding to pageXString and
			         * pageYString. Don't forget that y-values
			         * grow upwards for Postscript! */
  char *pageWidthString;	/* Printed width of output. */
  char *pageHeightString;	/* Printed height of output. */
  double scale;		        /* Scale factor for conversion: each pixel
			         * maps into this many points. */
  Tk_Anchor pageAnchor;	        /* How to anchor bbox on Postscript page. */
  int rotate;			/* Non-zero means output should be rotated
				 * on page (landscape mode). */
  char *fontVar;		/* If non-NULL, gives name of global variable
				 * containing font mapping information.
				 * Malloc'ed. */
  char *colorVar;		/* If non-NULL, give name of global variable
				 * containing color mapping information.
				 * Malloc'ed. */
  char *colorMode;		/* Mode for handling colors:  "monochrome",
				 * "gray", or "color".  Malloc'ed. */
  int colorLevel;		/* Numeric value corresponding to colorMode:
				 * 0 for mono, 1 for gray, 2 for color. */
  char *fileName;		/* Name of file in which to write Postscript;
				 * NULL means return Postscript info as
				 * result. Malloc'ed. */
  char *channelName;		/* If -channel is specified, the name of
                                 * the channel to use. */
  Tcl_Channel chan;		/* Open channel corresponding to fileName. */
  Tcl_HashTable fontTable;	/* Hash table containing names of all font
				 * families used in output.  The hash table
				 * values are not used. */
  int prepass;		        /* Non-zero means that we're currently in
			         * the pre-pass that collects font information,
			         * so the Postscript generated isn't
			         * relevant. */
} TkPostscriptInfo;

int Frame::postscriptProc(int prepass)
{
  TkPostscriptInfo* psInfo = ((TkCanvas*)canvas)->psInfoPtr;

  if (!visible)
    return TCL_OK;

  if (prepass)
    return TCL_OK;

  char buf[256];
  Vector org = psOrigin();

  // Image

  if (fits) {

    // save graphics state

    Tcl_AppendResult(interp, "gsave\n", NULL);

    float scale = psInfo->scale * psResolution / 72.;
    int width = (int)(options->width * scale);
    int height = (int)(options->height * scale);

    // scale

    {
      ostrstream str(buf,256);
      str << org[0] << ' ' << org[1] << " translate" << endl
	  << 1/scale << ' ' << 1/scale << " scale" << endl
	  << ends;
      Tcl_AppendResult(interp, buf, NULL);
    }

    switch (psLevel) {
    case 1:
      switch (psColorSpace) {
      case BW:
      case GRAY:
	psLevel1(GRAY, width, height, scale);
	break;
      case RGB:
      case CMYK:
	psLevel1(RGB, width, height, scale);
	break;
      }
      break;
    case 2:
      psLevel2(psColorSpace, width, height, scale);
      break;
    }

    // restore gstate

    Tcl_AppendResult(interp, "grestore\n", NULL);
  }

  // Markers & Contours

  // clip path
  // I don't know why the height needs to be reduced by one, however, it works

  {
    ostrstream str(buf,256);
    str << org[0] << ' ' << org[1] << ' ' 
	<< options->width << ' ' << options->height-1
	<< " rectclip" << endl << ends;
    Tcl_AppendResult(interp, buf, NULL);
  }

  switch (psLevel) {
  case 1:
    switch (psColorSpace) {
    case BW:
    case GRAY:
      psMarkers(GRAY);
      if (contour)
	contour->ps(GRAY);
      break;
    case RGB:
    case CMYK:
      psMarkers(RGB);
      if (contour)
	contour->ps(RGB);
      break;
    }
    break;
  case 2:
    psMarkers(psColorSpace);
    if (contour)
      contour->ps(psColorSpace);
    break;
  }

  return TCL_OK;
}

void Frame::psLevel1(PSColorSpace mode, int width, int height, float scale)
{
  if (fitsCount == 1)
    psLevel1Single(mode, width, height, scale);
  else
    psLevel1Mosaic(mode, width, height, scale);
}

void Frame::psLevel1Head(PSColorSpace mode, int width, int height)
{
  char buf[512];
  ostrstream str(buf,512);

  switch (mode) {
  case BW:
  case GRAY:
    str << "/picstr " << dec << width << " string def" << endl
	<< width << ' ' << height << " 8 matrix" << endl
	<< "{currentfile picstr readhexstring pop}" << endl
	<< "image" << endl;
    break;
  case RGB:
  case CMYK:
    str << "/picstr " << dec << width*3 << " string def" << endl
	<< width << ' ' << height << " 8 matrix" << endl
	<< "{currentfile picstr readhexstring pop}" << endl
	<< "false 3 colorimage" << endl;
    break;
  }

  str << ends;

  Tcl_AppendResult(interp, buf, NULL);
}

void Frame::psLevel1Single(PSColorSpace mode, int width, int height, float scale)
{
  psLevel1Head(mode, width, height);
  NoCompressAsciiHex filter;
  
  // size = 2 char per byte * 3 bytes per pixel + estimate of endls + endl + ends

  int size = (int)(width*3*2*(1 + 1./80.) +.5) + 2;
  char buf[size];

  int srcWidth = fits->width();
  int srcHeight = fits->height();

  int length = colorScale->size() - 1;
  const unsigned char* table = colorScale->psColors();

  Matrix m = (fits->getDataToRef() * psMatrix(scale, width, height)).invert();
  double ll = fits->getLowDouble();
  double hh = fits->getHighDouble();
  double diff = hh - ll;

  for (int j=height-1; j>=0; j--) {
    ostrstream str(buf,size);

    for (int i=0; i<width; i++) {
      unsigned char red = 255;
      unsigned char green = 255;
      unsigned char blue = 255;
	    
      Vector s = Vector(i,j) * m;
      if (s[0] >= 0 && s[0] < srcWidth && s[1] >= 0 && s[1] < srcHeight) {
	register double value = fits->getValueDouble(((int)s[1])*srcWidth + (int)s[0]);
	
	if (!isnand(value))
	  if (value <= ll) {
	    blue = table[0];
	    green = table[1];
	    red = table[2];
	  }
	  else if (value >= hh) {
	    blue = table[length*3];
	    green = table[length*3+1];
	    red = table[length*3+2];
	  }
	  else {
	    int l = (int)(((value - ll)/diff * length) + .5);
	    blue = table[l*3];
	    green = table[l*3+1];
	    red = table[l*3+2];
	  }
      }

      switch (mode) {
      case BW:
      case GRAY:
	filter << RGB2Gray(red, green, blue);
	break;
      case RGB:
      case CMYK:
	filter << red << green << blue;
	break;
      }
      str << filter;
    }
    str << ends;
    Tcl_AppendResult(interp, buf, NULL);
  }

  ostrstream str(buf,size);
  filter.flush(str);
  str << endl << ends;
  Tcl_AppendResult(interp, buf, NULL);
}

void Frame::psLevel1Mosaic(PSColorSpace mode, int width, int height, float scale)
{
  psLevel1Head(mode, width, height);

  NoCompressAsciiHex filter;

  // size = 2 char per byte * 3 bytes per pixel + estimate of endls + endl + ends

  int size = (int)(width*3*2*(1 + 1./80.) +.5) + 2;
  char buf[size];

  int length = colorScale->size() - 1;
  const unsigned char* table = colorScale->psColors();

  Matrix m;
  int srcWidth;
  int srcHeight;

  double ll;
  double hh;
  double diff;

  int found = 0;
  FitsImage* current = fits;
  FitsImage* ptr;
  Matrix m1 = psMatrix(scale, width, height);

  for (int j=height-1; j>=0; j--) {
    ostrstream str(buf,size);

    for (int i=0; i<width; i++) {
      unsigned char red = 255;
      unsigned char green = 255;
      unsigned char blue = 255;
	    
    tryagain:
      if (!found) {
	ptr = current;
	current = current->next();

	m = (ptr->getDataToRef() * m1).invert();

	srcWidth = ptr->width();
	srcHeight = ptr->height();

	ll = ptr->getLowDouble();
	hh = ptr->getHighDouble();
	diff = hh - ll;
      }

      Vector s = Vector(i,j) * m;

      if (s[0] >= 0 && s[0] < srcWidth && s[1] >= 0 && s[1] < srcHeight) {
	register double value = ptr->getValueDouble(((int)s[1])*srcWidth + (int)s[0]);
	
	if (!isnand(value))
	  if (value <= ll) {
	    blue = table[0];
	    green = table[1];
	    red = table[2];
	  }
	  else if (value >= hh) {
	    blue = table[length*3];
	    green = table[length*3+1];
	    red = table[length*3+2];
	  }
	  else {
	    int l = (int)(((value - ll)/diff * length) + .5);
	    blue = table[l*3];
	    green = table[l*3+1];
	    red = table[l*3+2];
	  }
	found = 1;
      }
      else {
	found = 0;
	if (current)
	  goto tryagain;
	else
	  current = fits;
      }

      switch (mode) {
      case BW:
      case GRAY:
	filter << RGB2Gray(red, green, blue);
	break;
      case RGB:
      case CMYK:
	filter << red << green << blue;
	break;
      }
      str << filter;
    }
    str << ends;
    Tcl_AppendResult(interp, buf, NULL);
  }

  ostrstream str(buf,size);
  filter.flush(str);
  str << endl << ends;
  Tcl_AppendResult(interp, buf, NULL);
}

void Frame::psLevel2(PSColorSpace mode, int width, int height, float scale)
{
  if (colorScale->size() < 256) {
    if (fitsCount == 1)
      psLevel2SingleIndex(mode, width, height, scale);
    else
      psLevel2MosaicIndex(mode, width, height, scale);
  }
  else {
    if (fitsCount == 1)
      psLevel2Single(mode, width, height, scale);
    else
      psLevel2Mosaic(mode, width, height, scale);
  }
}

void Frame::psLevel2Head(PSColorSpace mode, int width, int height)
{
  char buf[4096];
  ostrstream str(buf,4096);

  switch (mode) {
  case BW:
  case GRAY:
    str << "/DeviceGray setcolorspace" << endl;
    break;
  case RGB:
    str << "/DeviceRGB setcolorspace" << endl;
    break;
  case CMYK:
    str << "/DeviceCMYK setcolorspace" << endl;
    break;
  }

  str << "<<" << endl
      << "/ImageType 1" << endl
      << "/Width " << dec << width << endl
      << "/Height " << dec << height << endl
      << "/BitsPerComponent 8" << endl;

  switch (mode) {
  case BW:
  case GRAY:
    str << "/Decode [0 1]" << endl;
    break;
  case RGB:
    str << "/Decode [0 1 0 1 0 1]" << endl;
    break;
  case CMYK:
    str << "/Decode [0 1 0 1 0 1 0 1]" << endl;
    break;
  }

  if (psInterpolate)
    str << "/Interpolate true" << endl;

  str << "/ImageMatrix matrix" << endl
      << "/DataSource currentfile" << endl
      << "/ASCII85Decode filter" << endl
      << "/RunLengthDecode filter" << endl
      << ">>" << endl
      << "image" << endl
      << ends;

  Tcl_AppendResult(interp, buf, NULL);
}

void Frame::psLevel2Single(PSColorSpace mode, int width, int height, float scale)
{
  psLevel2Head(mode, width, height);
  RLEAscii85 filter;

  // size = 2 char per byte * 4 chars per pixel + estimate of endls + endl + ends

  int size = (int)(width*4*2*(1 + 1./80.) +.5) + 2;
  char buf[size];
	    
  int srcWidth = fits->width();
  int srcHeight = fits->height();

  int length = colorScale->size() - 1;
  const unsigned char* table = colorScale->psColors();

  Matrix m = (fits->getDataToRef() * psMatrix(scale, width, height)).invert();
  double ll = fits->getLowDouble();
  double hh = fits->getHighDouble();
  double diff = hh - ll;

  for (int j=height-1; j>=0; j--) {
    ostrstream str(buf,size);

    for (int i=0; i<width; i++) {
      unsigned char red = 255;
      unsigned char green = 255;
      unsigned char blue = 255;
	    
      Vector s = Vector(i,j) * m;
      if (s[0] >= 0 && s[0] < srcWidth && s[1] >= 0 && s[1] < srcHeight) {
	register double value = fits->getValueDouble(((int)s[1])*srcWidth + (int)s[0]);
	
	if (!isnand(value))
	  if (value <= ll) {
	    blue = table[0];
	    green = table[1];
	    red = table[2];
	  }
	  else if (value >= hh) {
	    blue = table[length*3];
	    green = table[length*3+1];
	    red = table[length*3+2];
	  }
	  else {
	    int l = (int)(((value - ll)/diff * length) + .5);
	    blue = table[l*3];
	    green = table[l*3+1];
	    red = table[l*3+2];
	  }
      }

      switch (mode) {
      case BW:
      case GRAY:
	filter << RGB2Gray(red, green, blue);
	break;
      case RGB:
	filter << red << green << blue;
	break;
      case CMYK:
	{
	  unsigned char cyan, magenta, yellow, black;
	  RGB2CMYK(red, green, blue, &cyan, &magenta, &yellow, &black);
	  filter << cyan << magenta << yellow << black;
	}
	break;
      }
      str << filter;
    }
    str << ends;
    Tcl_AppendResult(interp, buf, NULL);
  }

  ostrstream str(buf,size);
  filter.flush(str);
  str << endl << "~>" << endl << ends;
  Tcl_AppendResult(interp, buf, NULL);
}

void Frame::psLevel2Mosaic(PSColorSpace mode, int width, int height, float scale)
{
  psLevel2Head(mode, width, height);
  RLEAscii85 filter;

  // size = 2 char per byte * 4 bytes per pixel + estimate of endls + endl + ends

  int size = (int)(width*4*2*(1 + 1./80.) +.5) + 2;
  char buf[size];

  int length = colorScale->size() - 1;
  const unsigned char* table = colorScale->psColors();

  Matrix m;
  int srcWidth;
  int srcHeight;

  double ll;
  double hh;
  double diff;

  int found = 0;
  FitsImage* current = fits;
  FitsImage* ptr;
  Matrix m1 = psMatrix(scale, width, height);

  for (int j=height-1; j>=0; j--) {
    ostrstream str(buf,size);

    for (int i=0; i<width; i++) {
      unsigned char red = 255;
      unsigned char green = 255;
      unsigned char blue = 255;

    tryagain:
      if (!found) {
	ptr = current;
	current = current->next();

	m = (ptr->getDataToRef() * m1).invert();

	srcWidth = ptr->width();
	srcHeight = ptr->height();

	ll = ptr->getLowDouble();
	hh = ptr->getHighDouble();
	diff = hh - ll;
      }

      Vector s = Vector(i,j) * m;

      if (s[0] >= 0 && s[0] < srcWidth && s[1] >= 0 && s[1] < srcHeight) {
	register double value = ptr->getValueDouble(((int)s[1])*srcWidth + (int)s[0]);
	
	if (!isnand(value))
	  if (value <= ll) {
	    blue = table[0];
	    green = table[1];
	    red = table[2];
	  }
	  else if (value >= hh) {
	    blue = table[length*3];
	    green = table[length*3+1];
	    red = table[length*3+2];
	  }
	  else {
	    int l = (int)(((value - ll)/diff * length) + .5);
	    blue = table[l*3];
	    green = table[l*3+1];
	    red = table[l*3+2];
	  }
	found = 1;
      }
      else {
	found = 0;
	if (current)
	  goto tryagain;
	else
	  current = fits;
      }

      switch (mode) {
      case BW:
      case GRAY:
	filter << RGB2Gray(red, green, blue);
	break;
      case RGB:
	filter << red << green << blue;
	break;
      case CMYK:
	{
	  unsigned char cyan, magenta, yellow, black;
	  RGB2CMYK(red, green, blue, &cyan, &magenta, &yellow, &black);
	  filter << cyan << magenta << yellow << black;
	}
	break;
      }
      str << filter;
    }
    str << ends;
    Tcl_AppendResult(interp, buf, NULL);
  }

  ostrstream str(buf,size);
  filter.flush(str);
  str << endl << "~>" << endl << ends;
  Tcl_AppendResult(interp, buf, NULL);
}

void Frame::psLevel2HeadIndex(PSColorSpace mode, int width, int height)
{
  char buf[4096];
  ostrstream str(buf,4096);
  const unsigned short white = 255;

  switch (mode) {
  case BW:
  case GRAY:
    {
      str << "[/Indexed /DeviceGray 255" << endl 
	  << '<' << endl;

      int count = 0;
      for (int i=0; i<255; i++) {
	unsigned short value = 0;

	for (int j=0; j<colorCount; j++)
	  if (indexCells[j]==i) {
	    value = RGB2Gray(colorCells[j*3+2], colorCells[j*3+1], colorCells[j*3]);
	    break;
	  }

	str << hex << setfill('0') << setw(2) << value << ' ';

	if (++count == 16) {
	  str << endl;
	  count = 0;
	}
      }

      // set last location to white
      str << hex << setfill('0') << setw(2) << white << endl;

    }
    break;
  case RGB:
    {
      str << "[/Indexed /DeviceRGB 255" << endl 
	  << '<' << endl;

      int count = 0;
      for (int i=0; i<255; i++) {
	unsigned short red = 0;
	unsigned short green = 0;
	unsigned short blue = 0;

	for (int j=0; j<colorCount; j++)
	  if (indexCells[j]==i) {
	    red = colorCells[j*3+2];
	    green = colorCells[j*3+1];
	    blue = colorCells[j*3];
	    break;
	  }

	str << hex << setfill('0') << setw(2) << red 
	    << hex << setfill('0') << setw(2) << green
	    << hex << setfill('0') << setw(2) << blue
	    << ' ';

	if (++count == 8) {
	  str << endl;
	  count = 0;
	}
      }

      // set last location to white
      str << hex << setfill('0') << setw(2) << white
	  << hex << setfill('0') << setw(2) << white
	  << hex << setfill('0') << setw(2) << white
	  << ' ' << endl;

    }
    break;
  case CMYK:
    {
      str << "[/Indexed /DeviceCMYK 255" << endl 
	  << '<' << endl;

      int count = 0;
      for (int i=0; i<255; i++) {
	unsigned char c = 0;
	unsigned char m = 0;
	unsigned char y = 0;
	unsigned char b = 255;

	for (int j=0; j<colorCount; j++)
	  if (indexCells[j]==i) {
	    RGB2CMYK(colorCells[j*3+2], colorCells[j*3+1], colorCells[j*3],
		     &c, &m, &y, &b);
	    break;
	  }

	unsigned short cyan = c;
	unsigned short magenta = m;
	unsigned short yellow = y;
	unsigned short black = b;

	str << hex << setfill('0') << setw(2) << cyan 
	    << hex << setfill('0') << setw(2) << magenta
	    << hex << setfill('0') << setw(2) << yellow
	    << hex << setfill('0') << setw(2) << black
	    << ' ';

	if (++count == 8) {
	  str << endl;
	  count = 0;
	}
      }

      // set last location to white
      str << hex << setfill('0') << setw(2) << 0
	  << hex << setfill('0') << setw(2) << 0
	  << hex << setfill('0') << setw(2) << 0
	  << hex << setfill('0') << setw(2) << 0
	  << ' ' << endl;
    }
    break;
  }

  str << '>' << endl
      << "] setcolorspace" << endl
      << "<<" << endl
      << "/ImageType 1" << endl
      << "/Width " << dec << width << endl
      << "/Height " << dec << height << endl
      << "/BitsPerComponent 8" << endl
      << "/Decode [0 255]" << endl;

  if (psInterpolate)
    str << "/Interpolate true" << endl;

  str << "/ImageMatrix matrix" << endl
      << "/DataSource currentfile" << endl
      << "/ASCII85Decode filter" << endl
      << "/RunLengthDecode filter" << endl
      << ">>" << endl
      << "image" << endl
      << ends;

  Tcl_AppendResult(interp, buf, NULL);
}

void Frame::psLevel2SingleIndex(PSColorSpace mode, int width, int height, float scale)
{
  psLevel2HeadIndex(mode, width, height);
  RLEAscii85 filter;

  // size = 2 char per byte + estimate of endls + endl + ends

  int size = (int)(width*2*(1 + 1./80.) +.5) + 2;
  char buf[size];
	    
  int srcWidth = fits->width();
  int srcHeight = fits->height();

  int length = colorScale->size() - 1;
  const unsigned short* table = colorScale->psIndex();

  Matrix m = (fits->getDataToRef() * psMatrix(scale, width, height)).invert();
  double ll = fits->getLowDouble();
  double hh = fits->getHighDouble();
  double diff = hh - ll;

  for (int j=height-1; j>=0; j--) {
    ostrstream str(buf,size);

    for (int i=0; i<width; i++) {
      Vector s = Vector(i,j) * m;
      if (s[0] >= 0 && s[0] < srcWidth && s[1] >= 0 && s[1] < srcHeight) {
	register double value = fits->getValueDouble(((int)s[1])*srcWidth + (int)s[0]);
	
	if (!isnand(value))
	  if (value <= ll)
	    filter << table[0];
	  else if (value >= hh)
	    filter << table[length];
	  else
	    filter << table[(int)(((value - ll)/diff * length) + .5)];
	else
	  filter << 255;
      }
      else
	filter << 255;

      str << filter;
    }
    str << ends;
    Tcl_AppendResult(interp, buf, NULL);
  }

  ostrstream str(buf,size);
  filter.flush(str);
  str << endl << "~>" << endl << ends;
  Tcl_AppendResult(interp, buf, NULL);
}

void Frame::psLevel2MosaicIndex(PSColorSpace mode, int width, int height, float scale)
{
  psLevel2HeadIndex(mode, width, height);
  RLEAscii85 filter;

  // size = 2 char per byte * 3 bytes per pixel + estimate of endls + endl + ends

  int size = (int)(width*3*2*(1 + 1./80.) +.5) + 2;
  char buf[size];

  int length = colorScale->size() - 1;
  const unsigned short* table = colorScale->psIndex();

  Matrix m;
  int srcWidth;
  int srcHeight;

  double ll;
  double hh;
  double diff;

  int found = 0;
  FitsImage* current = fits;
  FitsImage* ptr;
  Matrix m1 = psMatrix(scale, width, height);

  for (int j=height-1; j>=0; j--) {
    ostrstream str(buf,size);

    for (int i=0; i<width; i++) {
    tryagain:
      if (!found) {
	ptr = current;
	current = current->next();

	m = (ptr->getDataToRef() * m1).invert();

	srcWidth = ptr->width();
	srcHeight = ptr->height();

	ll = ptr->getLowDouble();
	hh = ptr->getHighDouble();
	diff = hh - ll;
      }

      Vector s = Vector(i,j) * m;

      if (s[0] >= 0 && s[0] < srcWidth && s[1] >= 0 && s[1] < srcHeight) {
	register double value = ptr->getValueDouble(((int)s[1])*srcWidth + (int)s[0]);
	
	if (!isnand(value))
	  if (value <= ll)
	    filter << table[0];
	  else if (value >= hh)
	    filter << table[length];
	  else
	    filter << table[(int)(((value - ll)/diff * length) + .5)];

	found = 1;
      }
      else {
	found = 0;
	if (current)
	  goto tryagain;
	else {
	  filter << 255;
	  current = fits;
	}
      }

      str << filter;
    }
    str << ends;
    Tcl_AppendResult(interp, buf, NULL);
  }

  ostrstream str(buf,size);
  filter.flush(str);
  str << endl << "~>" << endl << ends;
  Tcl_AppendResult(interp, buf, NULL);
}

Matrix Frame::psMatrix(float psZoom, int width, int height)
{
  Vector center = getCenter();
  return 
    Translate(-center) *
    FlipY() *
    Translate(center) *

    Translate(-cursor) *                // translate to cursor position
    Rotate(rotation) *                  // rotate about cursor position
    Rotate(wcsRotation) *               // rotate about center position
    orientationMatrix *                 // flip x/y axis about cursor position
    wcsOrientationMatrix *              // flip x/y axis about center
    Scale(zoom) *                       // scale about cursor position
    Scale(wcsZoom) *                    // scale about wcs scale
    Scale(psZoom) *

    Translate(Vector(width,height)/2.);
}

Vector Frame::psOrigin()
{
  // compute coordinates of lower-left corner of the image
  // taking into account the anchor position for the image.
  // this mapping is different than in widget.c, due to postscript origin (y is flipped)

  double X = options->x;
  double Y = Tk_CanvasPsY(canvas, options->y);
     
  switch (options->anchor) {
  case TK_ANCHOR_NW:
    Y -= options->height;
    break;
  case TK_ANCHOR_N:
    X -= options->width/2.0;
    Y -= options->height;
    break;
  case TK_ANCHOR_NE:
    X -= options->width;
    Y -= options->height;
    break;
  case TK_ANCHOR_E:
    X -= options->width;
    Y -= options->height/2.0;
    break;
  case TK_ANCHOR_SE:
    X -= options->width;
    break;
  case TK_ANCHOR_S:
    X -= options->width/2.0;
    break;
  case TK_ANCHOR_SW:
    break;
  case TK_ANCHOR_W:
    Y -= options->height/2.0;
    break;
  case TK_ANCHOR_CENTER:
    X -= options->width/2.0;
    Y -= options->height/2.0;
    break;
  }

  return Vector(X,Y);
}

extern "C" {

int TkGetProlog(Tcl_Interp *interp)
{
  char prolog[] = 
{"
%%BeginProlog
50 dict begin

% This is a standard prolog for Postscript generated by Tk's canvas
% widget.
% SCCS: @(#) prolog.ps 1.7 96/07/08 17:52:14

% The definitions below just define all of the variables used in
% any of the procedures here.  This is needed for obscure reasons
% explained on p. 716 of the Postscript manual (Section H.2.7,
% 'Initializing Variables', in the section on Encapsulated Postscript).

/baseline 0 def
/stipimage 0 def
/height 0 def
/justify 0 def
/lineLength 0 def
/spacing 0 def
/stipple 0 def
/strings 0 def
/xoffset 0 def
/yoffset 0 def
/tmpstip null def

% Define the array ISOLatin1Encoding (which specifies how characters are
% encoded for ISO-8859-1 fonts), if it isn't already present (Postscript
% level 2 is supposed to define it, but level 1 doesn't).

systemdict /ISOLatin1Encoding known not {
    /ISOLatin1Encoding [
	/space /space /space /space /space /space /space /space
	/space /space /space /space /space /space /space /space
	/space /space /space /space /space /space /space /space
	/space /space /space /space /space /space /space /space
	/space /exclam /quotedbl /numbersign /dollar /percent /ampersand
	    /quoteright
	/parenleft /parenright /asterisk /plus /comma /minus /period /slash
	/zero /one /two /three /four /five /six /seven
	/eight /nine /colon /semicolon /less /equal /greater /question
	/at /A /B /C /D /E /F /G
	/H /I /J /K /L /M /N /O
	/P /Q /R /S /T /U /V /W
	/X /Y /Z /bracketleft /backslash /bracketright /asciicircum /underscore
	/quoteleft /a /b /c /d /e /f /g
	/h /i /j /k /l /m /n /o
	/p /q /r /s /t /u /v /w
	/x /y /z /braceleft /bar /braceright /asciitilde /space
	/space /space /space /space /space /space /space /space
	/space /space /space /space /space /space /space /space
	/dotlessi /grave /acute /circumflex /tilde /macron /breve /dotaccent
	/dieresis /space /ring /cedilla /space /hungarumlaut /ogonek /caron
	/space /exclamdown /cent /sterling /currency /yen /brokenbar /section
	/dieresis /copyright /ordfeminine /guillemotleft /logicalnot /hyphen
	    /registered /macron
	/degree /plusminus /twosuperior /threesuperior /acute /mu /paragraph
	    /periodcentered
	/cedillar /onesuperior /ordmasculine /guillemotright /onequarter
	    /onehalf /threequarters /questiondown
	/Agrave /Aacute /Acircumflex /Atilde /Adieresis /Aring /AE /Ccedilla
	/Egrave /Eacute /Ecircumflex /Edieresis /Igrave /Iacute /Icircumflex
	    /Idieresis
	/Eth /Ntilde /Ograve /Oacute /Ocircumflex /Otilde /Odieresis /multiply
	/Oslash /Ugrave /Uacute /Ucircumflex /Udieresis /Yacute /Thorn
	    /germandbls
	/agrave /aacute /acircumflex /atilde /adieresis /aring /ae /ccedilla
	/egrave /eacute /ecircumflex /edieresis /igrave /iacute /icircumflex
	    /idieresis
	/eth /ntilde /ograve /oacute /ocircumflex /otilde /odieresis /divide
	/oslash /ugrave /uacute /ucircumflex /udieresis /yacute /thorn
	    /ydieresis
    ] def
} if

% font ISOEncode font
% This procedure changes the encoding of a font from the default
% Postscript encoding to ISOLatin1.  It's typically invoked just
% before invoking 'setfont'.  The body of this procedure comes from
% Section 5.6.1 of the Postscript book.

/ISOEncode {
    dup length dict begin
	{1 index /FID ne {def} {pop pop} ifelse} forall
	/Encoding ISOLatin1Encoding def
	currentdict
    end

    % I'm not sure why it's necessary to use 'definefont' on this new
    % font, but it seems to be important; just use the name 'Temporary'
    % for the font.

    /Temporary exch definefont
} bind def

% StrokeClip
%
% This procedure converts the current path into a clip area under
% the assumption of stroking.  It's a bit tricky because some Postscript
% interpreters get errors during strokepath for dashed lines.  If
% this happens then turn off dashes and try again.

/StrokeClip {
    {strokepath} stopped {
	(This Postscript printer gets limitcheck overflows when) =
	(stippling dashed lines;  lines will be printed solid instead.) =
	[] 0 setdash strokepath} if
    clip
} bind def

% desiredSize EvenPixels closestSize
%
% The procedure below is used for stippling.  Given the optimal size
% of a dot in a stipple pattern in the current user coordinate system,
% compute the closest size that is an exact multiple of the device's
% pixel size.  This allows stipple patterns to be displayed without
% aliasing effects.

/EvenPixels {
    % Compute exact number of device pixels per stipple dot.
    dup 0 matrix currentmatrix dtransform
    dup mul exch dup mul add sqrt

    % Round to an integer, make sure the number is at least 1, and compute
    % user coord distance corresponding to this.
    dup round dup 1 lt {pop 1} if
    exch div mul
} bind def

% width height string StippleFill --
%
% Given a path already set up and a clipping region generated from
% it, this procedure will fill the clipping region with a stipple
% pattern.  'String' contains a proper image description of the
% stipple pattern and 'width' and 'height' give its dimensions.  Each
% stipple dot is assumed to be about one unit across in the current
% user coordinate system.  This procedure trashes the graphics state.

/StippleFill {
    % The following code is needed to work around a NeWSprint bug.

    /tmpstip 1 index def

    % Change the scaling so that one user unit in user coordinates
    % corresponds to the size of one stipple dot.
    1 EvenPixels dup scale

    % Compute the bounding box occupied by the path (which is now
    % the clipping region), and round the lower coordinates down
    % to the nearest starting point for the stipple pattern.  Be
    % careful about negative numbers, since the rounding works
    % differently on them.

    pathbbox
    4 2 roll
    5 index div dup 0 lt {1 sub} if cvi 5 index mul 4 1 roll
    6 index div dup 0 lt {1 sub} if cvi 6 index mul 3 2 roll

    % Stack now: width height string y1 y2 x1 x2
    % Below is a doubly-nested for loop to iterate across this area
    % in units of the stipple pattern size, going up columns then
    % across rows, blasting out a stipple-pattern-sized rectangle at
    % each position

    6 index exch {
	2 index 5 index 3 index {
	    % Stack now: width height string y1 y2 x y

	    gsave
	    1 index exch translate
	    5 index 5 index true matrix tmpstip imagemask
	    grestore
	} for
	pop
    } for
    pop pop pop pop pop
} bind def

% -- AdjustColor --
% Given a color value already set for output by the caller, adjusts
% that value to a grayscale or mono value if requested by the CL
% variable.

/AdjustColor {
    CL 2 lt {
	currentgray
	CL 0 eq {
	    .5 lt {0} {1} ifelse
	} if
	setgray
    } if
} bind def

% x y strings spacing xoffset yoffset justify stipple DrawText --
% This procedure does all of the real work of drawing text.  The
% color and font must already have been set by the caller, and the
% following arguments must be on the stack:
%
% x, y -	Coordinates at which to draw text.
% strings -	An array of strings, one for each line of the text item,
%		in order from top to bottom.
% spacing -	Spacing between lines.
% xoffset -	Horizontal offset for text bbox relative to x and y: 0 for
%		nw/w/sw anchor, -0.5 for n/center/s, and -1.0 for ne/e/se.
% yoffset -	Vertical offset for text bbox relative to x and y: 0 for
%		nw/n/ne anchor, +0.5 for w/center/e, and +1.0 for sw/s/se.
% justify -	0 for left justification, 0.5 for center, 1 for right justify.
% stipple -	Boolean value indicating whether or not text is to be
%		drawn in stippled fashion.  If text is stippled,
%		procedure StippleText must have been defined to call
%		StippleFill in the right way.
%
% Also, when this procedure is invoked, the color and font must already
% have been set for the text.

/DrawText {
    /stipple exch def
    /justify exch def
    /yoffset exch def
    /xoffset exch def
    /spacing exch def
    /strings exch def

    % First scan through all of the text to find the widest line.

    /lineLength 0 def
    strings {
	stringwidth pop
	dup lineLength gt {/lineLength exch def} {pop} ifelse
	newpath
    } forall

    % Compute the baseline offset and the actual font height.

    0 0 moveto (TXygqPZ) false charpath
    pathbbox dup /baseline exch def
    exch pop exch sub /height exch def pop
    newpath

    % Translate coordinates first so that the origin is at the upper-left
    % corner of the text's bounding box. Remember that x and y for
    % positioning are still on the stack.

    translate
    lineLength xoffset mul
    strings length 1 sub spacing mul height add yoffset mul translate

    % Now use the baseline and justification information to translate so
    % that the origin is at the baseline and positioning point for the
    % first line of text.

    justify lineLength mul baseline neg translate

    % Iterate over each of the lines to output it.  For each line,
    % compute its width again so it can be properly justified, then
    % display it.

    strings {
	dup stringwidth pop
	justify neg mul 0 moveto
	stipple {

	    % The text is stippled, so turn it into a path and print
	    % by calling StippledText, which in turn calls StippleFill.
	    % Unfortunately, many Postscript interpreters will get
	    % overflow errors if we try to do the whole string at
	    % once, so do it a character at a time.

	    gsave
	    /char (X) def
	    {
		char 0 3 -1 roll put
		currentpoint
		gsave
		char true charpath clip StippleText
		grestore
		char stringwidth translate
		moveto
	    } forall
	    grestore
	} {show} ifelse
	0 spacing neg translate
    } forall
} bind def

%%EndProlog
"};

  Tcl_AppendResult(interp, prolog, NULL);
  return TCL_OK;
}

}

