/*
 *      Copyrights:
 *
 *    	  Copyright (c) 1996, 1998 Smithsonian Astrophysical Observatory
 *
 *        Permission to use, copy, modify, distribute, and  sell  this
 *        software  and  its  documentation  for any purpose is hereby
 *        granted without  fee,  provided  that  the  above  copyright
 *        notice  appear  in  all  copies and that both that copyright
 *        notice and this permission notice appear in supporting docu-
 *        mentation,  and  that  the  name  of the  Smithsonian Astro-
 *	  physical Observatory not be used in advertising or publicity
 *  	  pertaining to distribution of the software without specific,
 *	  written  prior  permission.   The Smithsonian  Astrophysical 
 *        Observatory makes no representations about  the  suitability
 *	  of  this  software for any purpose.  It is provided  "as is"
 *	  without express or implied warranty.
 *        THE  SMITHSONIAN  ASTROPHYSICAL  OBSERVATORY  DISCLAIMS  ALL
 *	  WARRANTIES  WITH  REGARD  TO  THIS  SOFTWARE,  INCLUDING ALL
 *	  IMPLIED  WARRANTIES  OF  MERCHANTABILITY AND FITNESS, IN  NO
 *	  EVENT SHALL THE  SMITHSONIAN  ASTROPHYSICAL  OBSERVATORY  BE 
 *	  LIABLE FOR  ANY SPECIAL, INDIRECT  OR  CONSEQUENTIAL DAMAGES
 *	  OR ANY DAMAGES WHATSOEVER RESULTING FROM LOSS  OF USE,  DATA
 *	  OR  PROFITS, WHETHER IN AN ACTION OF CONTRACT, NEGLIGENCE OR
 *	  OTHER TORTIOUS ACTION, ARISING OUT OF OR IN CONNECTION  WITH
 *	  THE  USE OR PERFORMANCE OF THIS SOFTWARE.
 *
 */

#include <xpap.h>

extern char *optarg;
extern int optind;

typedef struct entrec{
  struct entrec *next;
  char *method;
  char *xclass;
  char *name;
  char *type;
  char *user;
} *Entry, EntryRec;

typedef struct reqrec{
  struct reqrec *next;
  int sock;
  FILE *fin;
  FILE *fout;
  Entry entry;
} *Req, ReqRec;

static char *helpbuf = "xpans commands:\nhelp\t\t\t\t# print this help message\nlist\t\t\t\t# list all entries\nlookup class:name [type]\t# lookup entries [of this type]";

static Req reqhead=NULL;
static int nentry = 0;
static int exconn=0;
static char *logfile=NULL;
static int sock;

static int LookupReq _PR((Req xreq, char *lbuf, int flag));
static int ListReq   _PR((Req xreq, int flag));
static void HelpReq   _PR((Req xreq, int flag));

/*
 *----------------------------------------------------------------------------
 *
 * Routine:	send_cb
 *
 * Purpose:	send callback for XPA access point
 *
 * Returns:	0 for success, -1 for failure
 *
 *----------------------------------------------------------------------------
 */
#ifdef ANSI_FUNC
static int 
send_cb (void *client_data, void *call_data, char *paramlist,
	 char **buf, int *len)
#else
static int send_cb(client_data, call_data, paramlist, buf, len)
     void *client_data;
     void *call_data;
     char *paramlist;
     char **buf;
     int *len;
#endif
{
  XPA xpa = (XPA)call_data;
  Req req;
  char tbuf[SZ_LINE*2];
  char ctmpl[SZ_LINE];
  char ntmpl[SZ_LINE];
  char type[SZ_LINE];
  char users[SZ_LINE];
  int wp=0;

  if( (req = (Req)calloc(1, sizeof(ReqRec))) == NULL )
    return(-1);
  req->sock = xpa_datafd(xpa);
  req->fin = fdopen(req->sock, "r");
  req->fout = fdopen(req->sock, "w");
  if( paramlist && *paramlist ){
    word(paramlist, tbuf, &wp );
    if( !strcmp(tbuf, "list") ){
      ListReq(req, 0);
    }
    else if( !strcmp(tbuf, "help") ){
      HelpReq(req, 0);
    }
    else{
      if( !strcmp(tbuf, "lookup") ){
	if( !word(paramlist, tbuf, &wp ) ){
	  strcpy(tbuf, "XPA$ERROR 'lookup' requires class:name [type]\n");
	  XPAPuts(req->fout, tbuf, XPALongTimeout());
	  return(-1);
	}
      }
      XPAParseName(tbuf, ctmpl, ntmpl, SZ_LINE);
      /* look for type */
      if( !word(paramlist, type, &wp ) )
	strcpy(type, XPA_ACLS);
      /* look for users */
      if( !word(paramlist, users, &wp ) )
	strcpy(type, "*");
      sprintf(tbuf, "%s %s %s %s", ctmpl, ntmpl, type, users);
      LookupReq(req, tbuf, 0);
    }
  }
  else {
    ListReq(req, 0);
  }
  fclose(req->fin);
  fclose(req->fout);
  close(req->sock);
  free(req);
  return(0);
}

/*
 *----------------------------------------------------------------------------
 *
 * Routine:	Log
 *
 * Purpose:	write all names to a backup log
 *
 * Returns:	none
 *
 *----------------------------------------------------------------------------
 */
#ifdef ANSI_FUNC
static void 
Log (void)
#else
static void Log()
#endif
{
  FILE *fp;
  Req req;
  Entry entry;

  if( !logfile )
    return;
  if( (fp = fopen(logfile, "w")) == NULL )
    return;
  for(req=reqhead; req!=NULL; req=req->next){
    for(entry=req->entry; entry!=NULL; entry=entry->next){
      fprintf(fp, "add %s %s %s %s %s\n",
	      entry->method, entry->xclass,
	      entry->name, entry->type, entry->user);
    }
  }
  fclose(fp);
}

/*
 *----------------------------------------------------------------------------
 *
 * Routine:	SplitArg
 *
 * Purpose:	split the specified argument by changing a ":" to a space
 *		splitting is done in place
 *
 * Returns:	none
 *
 *----------------------------------------------------------------------------
 */
#ifdef ANSI_FUNC
static void
SplitArg (char *buf, int arg)
#else
static void SplitArg(buf, arg)
     char *buf;
     int arg;
#endif
{
  int i;
  char *s;

  /* point to beginning of buffer */
  s = buf;
  /* skip over previous args */
  for(i=0; i<arg; i++){
    /* skip up to white space */
    while( *s && !isspace(*s) )
      s++;
    /* skip over white space to next arg */
    while( *s && isspace(*s) )
      s++;
  }
  /* we now are pointing at the arg in question.
     look for a ':' (up to next white space) and change to space */
  while( *s && !isspace(*s) ){
    if( *s == ':' ){
      *s = ' ';
      break;
    }
    else
      s++;
  }
}

/*
 *----------------------------------------------------------------------------
 *
 * Routine:	NewEntry
 *
 * Purpose:	allocate a new XPA entry
 *
 * Returns:	entry struct
 *
 *----------------------------------------------------------------------------
 */
#ifdef ANSI_FUNC
static Entry 
NewEntry (Req req,
	  char *method, char *xclass, char *name, char *type, char *user)
#else
static Entry NewEntry(req, method, xclass, name, type, user)
     Req req;
     char *method;
     char *xclass;
     char *name;
     char *type;
     char *user;
#endif
{
  Entry entry;
  Entry cur;

  if( (entry = (Entry)calloc(1, sizeof(EntryRec))) == NULL )
    return(NULL);

  /* fill in the blanks */
  entry->method = NewString(method);
  entry->xclass = NewString(xclass);
  entry->name = NewString(name);
  entry->type = NewString(type);
  entry->user = NewString(user);

  /* add this to end of the list */
  if( req->entry == NULL ){
    req->entry = entry;
  }
  else{
    for(cur=req->entry; cur->next!=NULL; cur=cur->next)
      ;
    cur->next = entry;
  }

  /* inc the total number of entries */
  nentry++;

  /* return the news */
  return(entry);
}

/*
 *----------------------------------------------------------------------------
 *
 * Routine:	FreeEntry
 *
 * Purpose:	free up an XPA entry
 *
 * Returns:	none
 *
 *----------------------------------------------------------------------------
 */
#ifdef ANSI_FUNC
static void 
FreeEntry (Req req, Entry entry)
#else
static void FreeEntry(req, entry)
     Req req;
     Entry entry;
#endif
{
  Entry cur;
  char tbuf[SZ_LINE];

  /* remove this entry from the list so it can't be found */
  if( entry == req->entry ){
    req->entry = req->entry->next;
  }
  else{
    for(cur=req->entry; cur!=NULL; cur=cur->next){
      if( entry == cur->next ){
	cur->next = entry->next;
	break;
      }
    }
  }

  /* perform method-specific cleanup */
  switch(XPAMethod(entry->method)){
  case XPA_UNIX:
    /* delete the unix socket files */
    unlink(entry->method);
    sprintf(tbuf, "%s_data", entry->method);
    unlink(tbuf);
    break;
  default:
    break;
  }

  /* now free this struct */
  if( entry->method )
    free(entry->method);
  if( entry->xclass )
    free(entry->xclass);
  if( entry->name )
    free(entry->name);
  if( entry->type )
    free(entry->type);
  if( entry->user )
    free(entry->user);
  free(entry);
}

/*
 *----------------------------------------------------------------------------
 *
 * Routine:	DelEntry
 *
 * Purpose:	Delete an XPA entry
 *
 * Returns:	none
 *
 *----------------------------------------------------------------------------
 */
#ifdef ANSI_FUNC
static void 
DelEntry (Req req, char *method)
#else
static void DelEntry(req, method)
     Req req;
     char *method;
#endif
{
  Entry cur, tcur;
  for(cur=req->entry; cur!=NULL; ){
    tcur = cur->next;
    if( ((method == NULL) || (*method == '\0')) ||
	!strcmp(cur->method, method)		){
      FreeEntry(req, cur);
    }
    cur = tcur;
  }
}

/*
 *----------------------------------------------------------------------------
 *
 * Routine:	NewReq
 *
 * Purpose:	allocate a new XPA request struct
 *
 * Returns:	Req struct
 *
 *----------------------------------------------------------------------------
 */
#ifdef ANSI_FUNC
static Req 
NewReq (int sock)
#else
static Req NewReq(sock)
     int sock;
#endif
{
  Req req;
  Req cur;

  if( (req = (Req)calloc(1, sizeof(ReqRec))) == NULL )
    return(NULL);

  /* fill in the blanks */
  req->sock = sock;
  req->fin = fdopen(req->sock, "r");
  req->fout = fdopen(req->sock, "w");

  /* add this to end of the list */
  if( reqhead == NULL ){
    reqhead = req;
  }
  else{
    for(cur=reqhead; cur->next!=NULL; cur=cur->next)
      ;
    cur->next = req;
  }
  /* return the news */
  return(req);
}

/*
 *----------------------------------------------------------------------------
 *
 * Routine:	FreeReq
 *
 * Purpose:	free up an XPA request entry
 *
 * Returns:	none
 *
 *----------------------------------------------------------------------------
 */
#ifdef ANSI_FUNC
static void 
FreeReq (Req req)
#else
static void FreeReq(req)
     Req req;
#endif
{
  Req cur;
  /* remove this entry from the list so it can't be found */
  if( req == reqhead ){
    reqhead = req->next;
  }
  else{
    for(cur=reqhead; cur!=NULL; cur=cur->next){
      if( req == cur->next ){
	cur->next = req->next;
	break;
      }
    }
  }
  /* close the communication channel */
  fclose(req->fin);
  fclose(req->fout);
  close(req->sock);
  /* now free this struct */
  free(req);
}

/*
 *----------------------------------------------------------------------------
 *
 * Routine:	AddReq
 *
 * Purpose:	add an XPA entry to the list
 *
 * Returns:	none
 *
 *----------------------------------------------------------------------------
 */
#ifdef ANSI_FUNC
static void 
AddReq (Req req, char *lbuf, int flag)
#else
static void AddReq(req, lbuf, flag)
     Req req;
     char *lbuf;
     int flag;
#endif
{
  int got;
  char xclass[SZ_LINE];
  char name[SZ_LINE];
  char method[SZ_LINE];
  char type[SZ_LINE];
  char user[SZ_LINE];
  char tbuf[SZ_LINE];

  SplitArg(lbuf, 1);
  got = sscanf(lbuf, "%s %s %s %s %s", method, xclass, name, type, user);
  if( got == 5 ){
    NewEntry(req, method, xclass, name, type, user);
    Log();
    if( flag )
      XPAPuts(req->fout, "XPA$OK\n", XPALongTimeout());
  }
  else{
    strcpy(tbuf, 
    "XPA$ERROR 'add' requires 4 args: ip:port class:name type user\n");
    XPAPuts(req->fout, tbuf, XPALongTimeout());
  }
}

/*
 *----------------------------------------------------------------------------
 *
 * Routine:	DelReq
 *
 * Purpose:	delete an XPA entry from the list
 *
 * Returns:	none
 *
 *----------------------------------------------------------------------------
 */
#ifdef ANSI_FUNC
static void 
DelReq (Req req, char *lbuf, int flag)
#else
static void DelReq(req, lbuf, flag)
     Req req;
     char *lbuf;
     int flag;
#endif
{
  int got;
  char tbuf[SZ_LINE];
  char method[SZ_LINE];

  if( lbuf != NULL ){
    if( (got = sscanf(lbuf, "%s", method)) == 1 ){
      /* free the specified entry */
      DelEntry(req, method);
      Log();
      if( flag )
	XPAPuts(req->fout, "XPA$OK\n", XPALongTimeout());
    }
    else{
      strcpy(tbuf, "XPA$ERROR 'del' requires 1 arg: ip:port\n");
      XPAPuts(req->fout, tbuf, XPALongTimeout());
    }
    /* if this is the last entry, delete the request struct as well */
    if( req->entry == NULL ){
      FreeReq(req);
    }
  }
  else{
    /* connection is closed -- free all entries for req, and delete req */
    DelEntry(req, NULL);
    Log();
    FreeReq(req);
  }
}

/*
 *----------------------------------------------------------------------------
 *
 * Routine:	LookupReq
 *
 * Purpose:	lookup a template in the XPA list
 *
 * Returns:	number of matched lookups
 *
 *----------------------------------------------------------------------------
 */
#ifdef ANSI_FUNC
static int 
LookupReq (Req xreq, char *lbuf, int flag)
#else
static int LookupReq(xreq, lbuf, flag)
     Req xreq;
     char *lbuf;
     int flag;
#endif
{
  int i;
  int got=0;
  int slen=1024;
  int *lens;
  char ctmpl[SZ_LINE];
  char ntmpl[SZ_LINE];
  char tbuf[SZ_LINE];
  char type[SZ_LINE];
  char users[SZ_LINE];
  char *user;
  char *usercopy;
  char **strings;
  char *tstring;
  Req req;
  Entry entry;

  SplitArg(lbuf, 0);
  if( sscanf(lbuf, "%s %s %s %s", ctmpl, ntmpl, type, users) == 4 ){
    strings = (char **)malloc(slen * sizeof(char *));
    lens = (int *)malloc(slen * sizeof(int));
    lens[0] = 0;
    for(req=reqhead; req!=NULL; req=req->next){
      for(entry=req->entry; entry!=NULL; entry=entry->next){
	if( (!strcmp(type, "*") || strpbrk(entry->type, type))   &&
	    tmatch(entry->xclass, ctmpl) 			 &&
	    tmatch(entry->name, ntmpl)  			 ){
	  if( !strcmp(users, "*") || !strcmp(users, entry->user) ){
	    user = entry->user;
	  }
	  else{
	    user = NULL;
	    usercopy = (char *)strdup(users);
	    for(user=(char *)strtok(usercopy, " :;,");
		user!=NULL;
		user=(char *)strtok(NULL," :;,")){
	      if ( !strcasecmp(user, entry->user) ){
		break;
	      }
	    }
	    if( usercopy )
	      free(usercopy);
	  }
	  if( user != NULL ){
	    sprintf(tbuf, "%s %s %s %s %s\n",
		    entry->xclass, entry->name,
		    entry->type, entry->method, entry->user);
	    if( got >= (slen-2) ){
	      slen *= 2;
	      strings = (char **)realloc(strings, slen * sizeof(char *));
	      lens = (int *)realloc(lens, slen * sizeof(int));
	    }
	    strings[got] = NewString(tbuf);
	    lens[got+1] = lens[got] + strlen(strings[got]);
	    got++;
	  }
	}
      }
    }
    if( flag ){
      strings[got] = NewString("XPA$OK\n");
      lens[got+1] = lens[got] + strlen(strings[got]);
      got++;
    }
    /* we have to avoid multiple writes in a row because tcp_delay
       buffers these (Nagle algorithm) and kills the performance */
    if( got > 0 ){
      tstring = (char *)calloc(lens[got]+1, sizeof(char));
      for(i=0; i<got; i++){
	if( strings[i] )
	  strcpy(&(tstring[lens[i]]), strings[i]);
      }
      XPAPutBuf(xreq->sock, tstring, lens[got], XPALongTimeout());
      /*      write(xreq->sock, tstring, lens[got]); */
      for(i=0; i<got; i++){
	if( strings[i] )
	  free(strings[i]);
      }
      if( tstring )
	free(tstring);
    }
    if( strings )
      free(strings);
    if( lens )
      free(lens);
  }
  else{
    strcpy(tbuf,
       "XPA$ERROR xpans 'lookup' requires 3 args: class:name type users\n");
    XPAPuts(xreq->fout, tbuf, XPALongTimeout());
  }
  return(got);
}

/*
 *----------------------------------------------------------------------------
 *
 * Routine:	HelpReq
 *
 * Purpose:	send help message
 *
 * Returns:	none
 *
 *----------------------------------------------------------------------------
 */
#ifdef ANSI_FUNC
static void
HelpReq (Req xreq, int flag)
#else
static void HelpReq(xreq, flag)
     Req xreq;
     int flag;
#endif
{
  XPAPuts(xreq->fout, helpbuf, XPALongTimeout());
  if( flag )
    XPAPuts(xreq->fout, "XPA$OK\n", XPALongTimeout());
}

/*
 *----------------------------------------------------------------------------
 *
 * Routine:	ListReq
 *
 * Purpose:	list all entries in the XPA list
 *
 * Returns:	number of entries listed
 *
 *----------------------------------------------------------------------------
 */
#ifdef ANSI_FUNC
static int 
ListReq (Req xreq, int flag)
#else
static int ListReq(xreq, flag)
     Req xreq;
     int flag;
#endif
{
  int got=0;
  char tbuf[SZ_LINE];
  Req req;
  Entry entry;

  for(req=reqhead; req!=NULL; req=req->next){
    for(entry=req->entry; entry!=NULL; entry=entry->next){
      sprintf(tbuf, "%s %s %s %s %s\n",
	      entry->xclass, entry->name,
	      entry->type, entry->method, entry->user);
      XPAPuts(xreq->fout, tbuf, XPALongTimeout());
      got++;
    }
  }
  if( flag )
    XPAPuts(xreq->fout, "XPA$OK\n", XPALongTimeout());
  return(got);
}

#ifdef ANSI_FUNC
void
usage (char *s)
#else
void usage(s)
     char *s;
#endif
{
    fprintf(stderr, "\n");
    fprintf(stderr, "usage: xpans [-h] [-e] [-p port] [-l log] \n");
    fprintf(stderr, "switches:\n");
    fprintf(stderr, "\t-h\tprint this message\n");
    fprintf(stderr, "\t-e\texit when there are no more XPA connections\n");
    fprintf(stderr, "\t-l file\tlog data base entries to specified file\n");
    fprintf(stderr, "\t-p port\tlisten for connections on specified port\n");
    fprintf(stderr, "\n");
    exit(1);
}

#ifdef ANSI_FUNC
int 
main (int argc, char **argv)
#else
main(argc, argv)
     int argc;
     char **argv;
#endif
{
  int c;
  char *s;
  char lbuf[SZ_LINE];
  char tbuf[SZ_LINE];
  char cmd[SZ_LINE];
  char file[SZ_LINE];
  char method[SZ_LINE];
  unsigned short port;
  int sock2;
  int width;
  int got;
  int mtype;
  int wp;
  int reuse_addr=1;
  int keep_alive=1;
  unsigned int ip, myip;
  int slen=sizeof(struct sockaddr_in);
  struct sockaddr_in sock_in;
  struct sockaddr_in sock_in2;
  struct sockaddr_un sock_un;
  struct sockaddr_in sock_un2;
  struct rlimit rlp;
  fd_set readfds;
  Req cur, tcur;
  XPA xpa;

  /* init the XPA environment */
  XPAInitEnv();

  /* Disable SIGPIPE so we do not die if the client dies.
   * Rather, we will get an EOF on reading or writing.
   */
  signal (SIGPIPE, SIG_IGN);

  /* get default ip and port */
  strcpy(method, XPANSMethod(0));
  mtype = XPAMethod(method);

  /* we want the args in the same order in which they arrived, and
     gnu getopt sometimes changes things without this */
  putenv("POSIXLY_CORRECT=true");

  /* process command line args */
  while ((c = getopt(argc, argv, "ef:hl:p:")) != -1){
    switch(c){
    case 'h':
      usage(argv[0]);
      exit(0);
    case 'e':
      exconn = 1;
      break;
    case 'f':
      mtype = XPA_UNIX;
      strcpy(file, optarg);
      strcpy(method, optarg);
    case 'l':
      logfile = optarg;
      break;
    case 'p':
      mtype = XPA_INET;
      port = atoi(optarg);
      strcpy(method, optarg);
      break;
    }
  }

  /* make sure we are running on the correct machine */
  switch(mtype){
  case XPA_INET:
    XPAParseIpPort(method, &ip, &port);
    if( (myip=gethostip(NULL)) != ip ){
      if( XPAVerbosity() > 1 )
	fprintf(stderr,
		"XPA$ERROR: XPA_NSINET variable expects xpans to run on %s\n",
		getiphost(myip));
      exit(1);
    }
    break;
  case XPA_UNIX:
    break;
  default:
    break;
  }

  /* set up communication method */
  switch(mtype){
  case XPA_INET:
    if( (sock = socket(AF_INET, SOCK_STREAM, 0)) < 0 ){
      if( XPAVerbosity() > 1 )
	perror("xpans socket()");
      exit(1);
    }
    setsockopt(sock, SOL_SOCKET, SO_KEEPALIVE,
	       (char *)&keep_alive, sizeof(keep_alive));
    memset((char *)&sock_in, 0, sizeof(sock_in));
    setsockopt(sock, SOL_SOCKET, SO_REUSEADDR,
	       (char *)&reuse_addr, sizeof(reuse_addr));
    sock_in.sin_family = AF_INET;
    sock_in.sin_addr.s_addr = htonl(gethostip(NULL));
    sock_in.sin_port = htons(port);
    /* bind to a port */
    if( bind(sock, (struct sockaddr *)&sock_in, sizeof(sock_in)) < 0 ){
      if( XPAVerbosity() > 1 )
	perror("xpans bind()");
      exit(1);
    }
    break;
  case XPA_UNIX:
    unlink(method);
    if( (sock = socket(AF_UNIX, SOCK_STREAM, 0)) < 0 ){
      if( XPAVerbosity() > 1 )
	perror("xpans socket()");
      exit(1);
    }
    setsockopt(sock, SOL_SOCKET, SO_KEEPALIVE,
	       (char *)&keep_alive, sizeof(keep_alive));
    memset((char *)&sock_un, 0, sizeof(sock_un));
    setsockopt(sock, SOL_SOCKET, SO_REUSEADDR,
	       (char *)&reuse_addr, sizeof(reuse_addr));
    sock_un.sun_family = AF_UNIX;
    strcpy(sock_un.sun_path, method);
    /* bind to a port */
    if( bind(sock, (struct sockaddr *)&sock_un, sizeof(sock_un)) < 0 ){
      if( XPAVerbosity() > 1 )
	perror("xpans bind()");
      exit(1);
    }
    break;
  default:
    break;
  }

  /* listen for connections */
  if( listen(sock, XPA_MAXLISTEN) < 0 ){
    if( XPAVerbosity() > 1 )
      perror("xpans listen()");
    exit(1);
  }
  /* make sure we close on exec */
  fcntl(sock, F_SETFD, FD_CLOEXEC);

  /* add an XPA access point for external processing */
  if( (xpa = XPANew(NULL, "xpans", helpbuf,
		    send_cb, NULL, "fillbuf=false",
		    NULL, NULL, NULL)) == NULL){
    if( XPAVerbosity() > 1 )
      fprintf(stderr, "XPA$ERROR: failed to create access point for xpans\n");
    exit(1);
  }

  /* init select parameters */
  getrlimit(RLIMIT_NOFILE, &rlp);
  width = rlp.rlim_cur;
  FD_ZERO(&readfds);
  /* enter processing loop */
  while( 1 ){
    /* add main listening socket */
    FD_SET(sock, &readfds);
    /* add request lines */
    for(got=0, cur=reqhead; cur!=NULL; cur=cur->next){
      FD_SET(cur->sock, &readfds);
      got++;
    }
    /* add XPA selections */
    XPAAddSelect(NULL, &readfds);
    /* if we once had entries but do not have then now, we might be done */
    if( exconn && !got && nentry  )
      exit(0);
    /* wait for next request */
    if( select(width, &readfds, NULL, NULL, NULL) ){
      /* process a new major request */
      if( FD_ISSET(sock, &readfds) ){
	/* new request */
	switch(mtype){
	case XPA_INET:
	  slen = sizeof(struct sockaddr_in);
	  if( (sock2=accept(sock, (struct sockaddr *)&sock_in2, &slen)) >=0 ){
	    /* make sure we close on exec */
	    fcntl(sock2, F_SETFD, FD_CLOEXEC);
	    NewReq(sock2);
	  }
	  break;
	case XPA_UNIX:
	  slen = sizeof(struct sockaddr_un);
	  if( (sock2=accept(sock, (struct sockaddr *)&sock_un2, &slen)) >=0 ){
	    /* make sure we close on exec */
	    fcntl(sock2, F_SETFD, FD_CLOEXEC);
	    NewReq(sock2);
	  }
	  break;
	default:
	  break;
	}
      }
      /* process an existing request line */
      for(got=0, cur=reqhead; cur!=NULL; ){
	tcur = cur->next;
	if( FD_ISSET(cur->sock, &readfds) ){
	  if( XPAGets(cur->fin, lbuf, SZ_LINE, XPAShortTimeout()) ){
	    /* process another request from this process */
	    wp = 0;
	    /* get first token: command */
	    if( !word(lbuf, cmd, &wp) ){
	      strcpy(tbuf, "XPA$ERROR no xpans command specified\n");
	      XPAPuts(cur->fout, tbuf, XPALongTimeout());
	    }
	    else{
	      /* skip white space */
	      while( isspace(lbuf[wp]) )
		wp++;
	      /* lookup the command */
	      if( !strcmp(cmd, "add") ){
		AddReq(cur, &(lbuf[wp]), 1);
	      }
	      else if( !strcmp(cmd, "del") ){
		DelReq(cur, &(lbuf[wp]), 1);
	      }
	      else if( !strcmp(cmd, "help") ){
		HelpReq(cur, 1);
	      }
	      else if( !strcmp(cmd, "list") ){
		ListReq(cur, 1);
	      }
	      else if( !strcmp(cmd, "lookup") ){
		LookupReq(cur, &(lbuf[wp]), 1);
	      }
	      else{
		sprintf(tbuf, "XPA$ERROR unknown xpans request: %s\n", cmd);
		XPAPuts(cur->fout, tbuf, XPALongTimeout());
	      }
	    }
	  }
	  else{
	    /* process dies: delete all entries associated with this sock */
	    DelReq(cur, NULL, 1);
	  }
	}
	cur = tcur;
      }
      /* process xpa requests */
      XPAProcessSelect(&readfds, 0);
      /* reset flag */
      FD_ZERO(&readfds);
    }
  }
 }
